#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright (c) 2014 Abhay Devasthale and Martin Raspaud

# Author(s):

#   Abhay Devasthale <abhay.devasthale@smhi.se>
#   Martin Raspaud <martin.raspaud@smhi.se>
#   Adam Dybbroe <adam.dybbroe@smhi.se>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Read a gac file.
Reads L1b GAC data from KLM series of satellites (NOAA-15 and later) and does most of the computations.
Format specification can be found here:
http://www.ncdc.noaa.gov/oa/pod-guide/ncdc/docs/klm/html/c8/sec83142-1.htm

"""

import numpy as np
from pygac.gac_reader import GACReader
import pygac.geotiepoints as gtp
import datetime
from pygac import gac_io
import logging
LOG = logging.getLogger(__name__)

#np.set_printoptions(threshold=np.nan)

# GAC header object

header = np.dtype([("data_set_creation_site_id", "S3"),
                   ("ascii_blank_=_x20", "S1"),
                   ("noaa_level_1b_format_version_number:", ">u2"),
                   ("noaa_level_1b_format_version_year", ">u2"),
                   ("noaa_level_1b_format_version_day_of_year", ">u2"),
                   ("reserved_for_logical_record_length", ">u2"),
                   ("reserved_for_block_size", ">u2"),
                   ("count_of_header_records", ">u2"),
                   ("zero_fill0", ">i2", (3, )),
                   ("data_set_name", "S42"),
                   ("processing_block_identification", "S8"),
                   ("noaa_spacecraft_identification_code", ">u2"),
                   ("instrument_id", ">u2"),
                   ("data_type_code", ">u2"),
                   ("tip_source_code", ">u2"),
                   ("start_of_data_set_day_count_starting_from_0_at_00h,_1_jan_1950",
                    ">u4"),
                   ("start_of_data_set_year", ">u2"),
                   ("start_of_data_set_day_of_year", ">u2"),
                   ("start_of_data_set_utc_time_of_day", ">u4"),
                   ("end_of_data_set_day_count_starting_from_0_at_00h,_1_jan_1950",
                    ">u4"),
                   ("end_of_data_set_year", ">u2"),
                   ("end_of_data_set_day_of_year", ">u2"),
                   ("end_of_data_set_utc_time_of_day", ">u4"),
                   ("year_of_last_cpids_update", ">u2"),
                   ("day_of_year_of_last_cpids_update", ">u2"),
                   ("zero_fill1", ">i2", (4, )),
                   # DATA SET QUALITY INDICATORS
                   ("instrument_status", ">u4"),
                   ("zero_fill2", ">i2"),
                   ("record_number_of_status_change", ">u2"),
                   ("second_instrument_status", ">u4"),
                   ("count_of_data_records", ">u2"),
                   ("count_of_calibrated,_earth_located_scan_lines", ">u2"),
                   ("count_of_missing_scan_lines", ">u2"),
                   ("count_of_data_gaps", ">u2"),
                   ("filler0", ">u2"),
                   ("count_of_pacs_detected_tip_parity_errors", ">u2"),
                   ("filler1", ">u2"),
                   ("time_sequence_error", ">u2"),
                   ("time_sequence_error_code", ">u2"),
                   ("socc_clock_update_indicator", ">u2"),
                   ("earth_location_error_indicator", ">u2"),
                   ("earth_location_error_code", ">u2"),
                   ("pacs_status_bit_field", ">u2"),
                   ("data_source", ">u2"),
                   ("zero_fill3", ">i4"),
                   ("reserved_for_the_ingester", "S8"),
                   ("reserved_for_decommutation", "S8"),
                   ("zero_fill4", ">i2", (5, )),
                   # CALIBRATION
                   ("ramp_auto_calibration_indicators_bit_field", ">u2"),
                   ("year_of_most_recent_solar_channel_calibration", ">u2"),
                   ("day_of_year_of_most_recent_solar_channel_calibration",
                    ">u2"),
                   ("primary_calibration_algorithm_id", ">u2"),
                   ("primary_calibration_algorithm_selected_options", ">u2"),
                   ("secondary_calibration_algorithm_id", ">u2"),
                   ("secondary_calibration_algorithm_selected_options", ">u2"),
                   ("ir_target_temperature_1_conversion_coefficient_1", ">i2"),
                   ("ir_target_temperature_1_conversion_coefficient_2", ">i2"),
                   ("ir_target_temperature_1_conversion_coefficient_3", ">i2"),
                   ("ir_target_temperature_1_conversion_coefficient_4", ">i2"),
                   ("ir_target_temperature_1_conversion_coefficient_5", ">i2"),
                   ("ir_target_temperature_1_conversion_coefficient_6", ">i2"),
                   ("ir_target_temperature_2_conversion_coefficient_1", ">i2"),
                   ("ir_target_temperature_2_conversion_coefficient_2", ">i2"),
                   ("ir_target_temperature_2_conversion_coefficient_3", ">i2"),
                   ("ir_target_temperature_2_conversion_coefficient_4", ">i2"),
                   ("ir_target_temperature_2_conversion_coefficient_5", ">i2"),
                   ("ir_target_temperature_2_conversion_coefficient_6", ">i2"),
                   ("ir_target_temperature_3_conversion_coefficient_1", ">i2"),
                   ("ir_target_temperature_3_conversion_coefficient_2", ">i2"),
                   ("ir_target_temperature_3_conversion_coefficient_3", ">i2"),
                   ("ir_target_temperature_3_conversion_coefficient_4", ">i2"),
                   ("ir_target_temperature_3_conversion_coefficient_5", ">i2"),
                   ("ir_target_temperature_3_conversion_coefficient_6", ">i2"),
                   ("ir_target_temperature_4_conversion_coefficient_1", ">i2"),
                   ("ir_target_temperature_4_conversion_coefficient_2", ">i2"),
                   ("ir_target_temperature_4_conversion_coefficient_3", ">i2"),
                   ("ir_target_temperature_4_conversion_coefficient_4", ">i2"),
                   ("ir_target_temperature_4_conversion_coefficient_5", ">i2"),
                   ("ir_target_temperature_4_conversion_coefficient_6", ">i2"),
                   ("zero_fill5", ">i4", (2, )),
                   # RADIANCE CONVERSION
                   ("ch_1_solar_filtered_irradiance_in_wavelength", ">i4"),
                   ("ch_1_equivalent_filter_width_in_wavelength", ">i4"),
                   ("ch_2_solar_filtered_irradiance_in_wavelength", ">i4"),
                   ("ch_2_equivalent_filter_width_in_wavelength", ">i4"),
                   ("ch_3a_solar_filtered_irradiance_in_wavelength", ">i4"),
                   ("ch_3a_equivalent_filter_width_in_wavelength", ">i4"),
                   ("ch_3b_central_wavenumber_(see", ">i4"),
                   ("ch_3b_constant_1", ">i4"),
                   ("ch_3b_constant_2", ">i4"),
                   ("ch_4_central_wavenumber", ">i4"),
                   ("ch_4_constant_1", ">i4"),
                   ("ch_4_constant_2", ">i4"),
                   ("ch_5_central_wavenumber", ">i4"),
                   ("ch_5_constant_1", ">i4"),
                   ("ch_5_constant_2", ">i4"),
                   ("zero_fill6", ">i4", (3, )),
                   # NAVIGATION
                   ("reference_ellipsoid_model_id", "S8"),
                   ("nadir_earth_location_tolerance", ">u2"),
                   ("earth_location_bit_field", ">u2"),
                   ("zero_fill7", ">i2"),
                   ("constant_roll_attitude_error", ">i2"),
                   ("constant_pitch_attitude_error", ">i2"),
                   ("constant_yaw_attitude_error", ">i2"),
                   ("epoch_year_for_orbit_vector", ">u2"),
                   ("day_of_epoch_year_for_orbit_vector", ">u2"),
                   ("epoch_utc_time_of_day_for_orbit_vector", ">u4"),
                   ("semi-major_axis", ">i4"),
                   ("eccentricity", ">i4"),
                   ("inclination", ">i4"),
                   ("argument_of_perigee", ">i4"),
                   ("right_ascension_of_the_ascending_node", ">i4"),
                   ("mean_anomaly", ">i4"),
                   ("position_vector_x_component", ">i4"),
                   ("position_vector_y_component", ">i4"),
                   ("position_vector_z_component", ">i4"),
                   ("velocity_vector_x-dot_component", ">i4"),
                   ("velocity_vector_y-dot_component", ">i4"),
                   ("velocity_vector_z-dot_component", ">i4"),
                   ("earth_sun_distance_ratio", ">u4"),
                   ("zero_fill8", ">i4", (4, )),
                   # ANALOG TELEMETRY CONVERSION
                   ("patch_temperature_conversion_coefficient_1", ">i2"),
                   ("patch_temperature_conversion_coefficient_2", ">i2"),
                   ("patch_temperature_conversion_coefficient_3", ">i2"),
                   ("patch_temperature_conversion_coefficient_4", ">i2"),
                   ("patch_temperature_conversion_coefficient_5", ">i2"),
                   ("reserved0", ">i2"),
                   ("patch_temperature_extended_conversion_coefficient_1",
                    ">i2"),
                   ("patch_temperature_extended_conversion_coefficient_2",
                    ">i2"),
                   ("patch_temperature_extended_conversion_coefficient_3",
                    ">i2"),
                   ("patch_temperature_extended_conversion_coefficient_4",
                    ">i2"),
                   ("patch_temperature_extended_conversion_coefficient_5",
                    ">i2"),
                   ("reserved1", ">i2"),
                   ("patch_power_conversion_coefficient_1", ">i2"),
                   ("patch_power_conversion_coefficient_2", ">i2"),
                   ("patch_power_conversion_coefficient_3", ">i2"),
                   ("patch_power_conversion_coefficient_4", ">i2"),
                   ("patch_power_conversion_coefficient_5", ">i2"),
                   ("reserved2", ">i2"),
                   ("radiator_temperature_conversion_coefficient_1", ">i2"),
                   ("radiator_temperature_conversion_coefficient_2", ">i2"),
                   ("radiator_temperature_conversion_coefficient_3", ">i2"),
                   ("radiator_temperature_conversion_coefficient_4", ">i2"),
                   ("radiator_temperature_conversion_coefficient_5", ">i2"),
                   ("reserved3", ">i2"),
                   ("blackbody_temperature_1_conversion_coefficient_1", ">i2"),
                   ("blackbody_temperature_1_conversion_coefficient_2", ">i2"),
                   ("blackbody_temperature_1_conversion_coefficient_3", ">i2"),
                   ("blackbody_temperature_1_conversion_coefficient_4", ">i2"),
                   ("blackbody_temperature_1_conversion_coefficient_5", ">i2"),
                   ("reserved4", ">i2"),
                   ("blackbody_temperature_2_conversion_coefficient_1", ">i2"),
                   ("blackbody_temperature_2_conversion_coefficient_2", ">i2"),
                   ("blackbody_temperature_2_conversion_coefficient_3", ">i2"),
                   ("blackbody_temperature_2_conversion_coefficient_4", ">i2"),
                   ("blackbody_temperature_2_conversion_coefficient_5", ">i2"),
                   ("reserved5", ">i2"),
                   ("blackbody_temperature_3_conversion_coefficient_1", ">i2"),
                   ("blackbody_temperature_3_conversion_coefficient_2", ">i2"),
                   ("blackbody_temperature_3_conversion_coefficient_3", ">i2"),
                   ("blackbody_temperature_3_conversion_coefficient_4", ">i2"),
                   ("blackbody_temperature_3_conversion_coefficient_5", ">i2"),
                   ("reserved6", ">i2"),
                   ("blackbody_temperature_4_conversion_coefficient_1", ">i2"),
                   ("blackbody_temperature_4_conversion_coefficient_2", ">i2"),
                   ("blackbody_temperature_4_conversion_coefficient_3", ">i2"),
                   ("blackbody_temperature_4_conversion_coefficient_4", ">i2"),
                   ("blackbody_temperature_4_conversion_coefficient_5", ">i2"),
                   ("reserved7", ">i2"),
                   ("electronics_current_conversion_coefficient_1", ">i2"),
                   ("electronics_current_conversion_coefficient_2", ">i2"),
                   ("electronics_current_conversion_coefficient_3", ">i2"),
                   ("electronics_current_conversion_coefficient_4", ">i2"),
                   ("electronics_current_conversion_coefficient_5", ">i2"),
                   ("reserved8", ">i2"),
                   ("motor_current_conversion_coefficient_1", ">i2"),
                   ("motor_current_conversion_coefficient_2", ">i2"),
                   ("motor_current_conversion_coefficient_3", ">i2"),
                   ("motor_current_conversion_coefficient_4", ">i2"),
                   ("motor_current_conversion_coefficient_5", ">i2"),
                   ("reserved9", ">i2"),
                   ("earth_shield_position_conversion_coefficient_1", ">i2"),
                   ("earth_shield_position_conversion_coefficient_2", ">i2"),
                   ("earth_shield_position_conversion_coefficient_3", ">i2"),
                   ("earth_shield_position_conversion_coefficient_4", ">i2"),
                   ("earth_shield_position_conversion_coefficient_5", ">i2"),
                   ("reserved10", ">i2"),
                   ("electronics_temperature_conversion_coefficient_1", ">i2"),
                   ("electronics_temperature_conversion_coefficient_2", ">i2"),
                   ("electronics_temperature_conversion_coefficient_3", ">i2"),
                   ("electronics_temperature_conversion_coefficient_4", ">i2"),
                   ("electronics_temperature_conversion_coefficient_5", ">i2"),
                   ("reserved11", ">i2"),
                   ("cooler_housing_temperature_conversion_coefficient_1",
                    ">i2"),
                   ("cooler_housing_temperature_conversion_coefficient_2",
                    ">i2"),
                   ("cooler_housing_temperature_conversion_coefficient_3",
                    ">i2"),
                   ("cooler_housing_temperature_conversion_coefficient_4",
                    ">i2"),
                   ("cooler_housing_temperature_conversion_coefficient_5",
                    ">i2"),
                   ("reserved12", ">i2"),
                   ("baseplate_temperature_conversion_coefficient_1", ">i2"),
                   ("baseplate_temperature_conversion_coefficient_2", ">i2"),
                   ("baseplate_temperature_conversion_coefficient_3", ">i2"),
                   ("baseplate_temperature_conversion_coefficient_4", ">i2"),
                   ("baseplate_temperature_conversion_coefficient_5", ">i2"),
                   ("reserved13", ">i2"),
                   ("motor_housing_temperature_conversion_coefficient_1",
                    ">i2"),
                   ("motor_housing_temperature_conversion_coefficient_2",
                    ">i2"),
                   ("motor_housing_temperature_conversion_coefficient_3",
                    ">i2"),
                   ("motor_housing_temperature_conversion_coefficient_4",
                    ">i2"),
                   ("motor_housing_temperature_conversion_coefficient_5",
                    ">i2"),
                   ("reserved14", ">i2"),
                   ("a/d_converter_temperature_conversion_coefficient_1",
                    ">i2"),
                   ("a/d_converter_temperature_conversion_coefficient_2",
                    ">i2"),
                   ("a/d_converter_temperature_conversion_coefficient_3",
                    ">i2"),
                   ("a/d_converter_temperature_conversion_coefficient_4",
                    ">i2"),
                   ("a/d_converter_temperature_conversion_coefficient_5",
                    ">i2"),
                   ("reserved15", ">i2"),
                   ("detector_#4_bias_voltage_conversion_coefficient_1",
                    ">i2"),
                   ("detector_#4_bias_voltage_conversion_coefficient_2",
                    ">i2"),
                   ("detector_#4_bias_voltage_conversion_coefficient_3",
                    ">i2"),
                   ("detector_#4_bias_voltage_conversion_coefficient_4",
                    ">i2"),
                   ("detector_#4_bias_voltage_conversion_coefficient_5",
                    ">i2"),
                   ("reserved16", ">i2"),
                   ("detector_#5_bias_voltage_conversion_coefficient_1",
                    ">i2"),
                   ("detector_#5_bias_voltage_conversion_coefficient_2",
                    ">i2"),
                   ("detector_#5_bias_voltage_conversion_coefficient_3",
                    ">i2"),
                   ("detector_#5_bias_voltage_conversion_coefficient_4",
                    ">i2"),
                   ("detector_#5_bias_voltage_conversion_coefficient_5",
                    ">i2"),
                   ("reserved17", ">i2"),
                   ("channel_3b_blackbody_view_conversion_coefficient_1",
                    ">i2"),
                   ("channel_3b_blackbody_view_conversion_coefficient_2",
                    ">i2"),
                   ("channel_3b_blackbody_view_conversion_coefficient_3",
                    ">i2"),
                   ("channel_3b_blackbody_view_conversion_coefficient_4",
                    ">i2"),
                   ("channel_3b_blackbody_view_conversion_coefficient_5",
                    ">i2"),
                   ("reserved18", ">i2"),
                   ("channel_4_blackbody_view_conversion_coefficient_1",
                    ">i2"),
                   ("channel_4_blackbody_view_conversion_coefficient_2",
                    ">i2"),
                   ("channel_4_blackbody_view_conversion_coefficient_3",
                    ">i2"),
                   ("channel_4_blackbody_view_conversion_coefficient_4",
                    ">i2"),
                   ("channel_4_blackbody_view_conversion_coefficient_5",
                    ">i2"),
                   ("reserved19", ">i2"),
                   ("channel_5_blackbody_view_conversion_coefficient_1",
                    ">i2"),
                   ("channel_5_blackbody_view_conversion_coefficient_2",
                    ">i2"),
                   ("channel_5_blackbody_view_conversion_coefficient_3",
                    ">i2"),
                   ("channel_5_blackbody_view_conversion_coefficient_4",
                    ">i2"),
                   ("channel_5_blackbody_view_conversion_coefficient_5",
                    ">i2"),
                   ("reserved20", ">i2"),
                   ("reference_voltage_conversion_coefficient_1", ">i2"),
                   ("reference_voltage_conversion_coefficient_2", ">i2"),
                   ("reference_voltage_conversion_coefficient_3", ">i2"),
                   ("reference_voltage_conversion_coefficient_4", ">i2"),
                   ("reference_voltage_conversion_coefficient_5", ">i2"),
                   ("reserved21", ">i2")])
# FILLER
#                  ("zero_fill9", ">i4", (4608 - 688, ))])


# video data object

scanline = np.dtype([("scan_line_number", ">u2"),
                     ("scan_line_year", ">u2"),
                     ("scan_line_day_of_year", ">u2"),
                     ("satellite_clock_drift_delta", ">i2"),
                     ("scan_line_utc_time_of_day", ">u4"),
                     ("scan_line_bit_field", ">u2"),
                     ("zero_fill0", ">i2", (5, )),
                     # QUALITY INDICATORS
                     ("quality_indicator_bit_field", ">u4"),
                     ("scan_line_quality_flags", ">u4"),
                     ("calibration_quality_flags", ">u2", (3, )),
                     ("count_of_bit_errors_in_frame_sync", ">u2"),
                     ("zero_fill1", ">i4", (2, )),
                     # CALIBRATION COEFFICIENTS
                     ("visible_operational_cal_ch_1_slope_1", ">i4"),
                     ("visible_operational_cal_ch_1_intercept_1", ">i4"),
                     ("visible_operational_cal_ch_1_slope_2", ">i4"),
                     ("visible_operational_cal_ch_1_intercept_2", ">i4"),
                     ("visible_operational_cal_ch_1_intersection", ">i4"),
                     ("visible_test_cal_ch_1_slope_1", ">i4"),
                     ("visible_test_cal_ch_1_intercept_1", ">i4"),
                     ("visible_test_cal_ch_1_slope_2", ">i4"),
                     ("visible_test_cal_ch_1_intercept_2", ">i4"),
                     ("visible_test_cal_ch_1_intersection", ">i4"),
                     ("visible_prelaunch_cal_ch_1_slope_1", ">i4"),
                     ("visible_prelaunch_cal_ch_1_intercept_1", ">i4"),
                     ("visible_prelaunch_cal_ch_1_slope_2", ">i4"),
                     ("visible_prelaunch_cal_ch_1_intercept_2", ">i4"),
                     ("visible_prelaunch_cal_ch_1_intersection", ">i4"),
                     ("visible_operational_cal_ch_2_slope_1", ">i4"),
                     ("visible_operational_cal_ch_2_intercept_1", ">i4"),
                     ("visible_operational_cal_ch_2_slope_2", ">i4"),
                     ("visible_operational_cal_ch_2_intercept_2", ">i4"),
                     ("visible_operational_cal_ch_2_intersection", ">i4"),
                     ("visible_test_cal_ch_2_slope_1", ">i4"),
                     ("visible_test_cal_ch_2_intercept_1", ">i4"),
                     ("visible_test_cal_ch_2_slope_2", ">i4"),
                     ("visible_test_cal_ch_2_intercept_2", ">i4"),
                     ("visible_test_cal_ch_2_intersection", ">i4"),
                     ("visible_prelaunch_cal_ch_2_slope_1", ">i4"),
                     ("visible_prelaunch_cal_ch_2_intercept_1", ">i4"),
                     ("visible_prelaunch_cal_ch_2_slope_2", ">i4"),
                     ("visible_prelaunch_cal_ch_2_intercept_2", ">i4"),
                     ("visible_prelaunch_cal_ch_2_intersection", ">i4"),
                     ("visible_operational_cal_ch_3a_slope_1", ">i4"),
                     ("visible_operational_cal_ch_3a_intercept_1", ">i4"),
                     ("visible_operational_cal_ch_3a_slope_2", ">i4"),
                     ("visible_operational_cal_ch_3a_intercept_2", ">i4"),
                     ("visible_operational_cal_ch_3a_intersection", ">i4"),
                     ("visible_test_cal_ch_3a_slope_1", ">i4"),
                     ("visible_test_cal_ch_3a_intercept_1", ">i4"),
                     ("visible_test_cal_ch_3a_slope_2", ">i4"),
                     ("visible_test_cal_ch_3a_intercept_2", ">i4"),
                     ("visible_test_cal_ch_3a_intersection", ">i4"),
                     ("visible_prelaunch_cal_ch_3a_slope_1", ">i4"),
                     ("visible_prelaunch_cal_ch_3a_intercept_1", ">i4"),
                     ("visible_prelaunch_cal_ch_3a_slope_2", ">i4"),
                     ("visible_prelaunch_cal_ch_3a_intercept_2", ">i4"),
                     ("visible_prelaunch_cal_ch_3a_intersection", ">i4"),
                     ("ir_operational_cal_ch_3b_coefficient_1", ">i4"),
                     ("ir_operational_cal_ch_3b_coefficient_2", ">i4"),
                     ("ir_operational_cal_ch_3b_coefficient_3", ">i4"),
                     ("ir_test_cal_ch_3b_coefficient_1", ">i4"),
                     ("ir_test_cal_ch_3b_coefficient_2", ">i4"),
                     ("ir_test_cal_ch_3b_coefficient_3", ">i4"),
                     ("ir_operational_cal_ch_4_coefficient_1", ">i4"),
                     ("ir_operational_cal_ch_4_coefficient_2", ">i4"),
                     ("ir_operational_cal_ch_4_coefficient_3", ">i4"),
                     ("ir_test_cal_ch_4_coefficient_1", ">i4"),
                     ("ir_test_cal_ch_4_coefficient_2", ">i4"),
                     ("ir_test_cal_ch_4_coefficient_3", ">i4"),
                     ("ir_operational_cal_ch_5_coefficient_1", ">i4"),
                     ("ir_operational_cal_ch_5_coefficient_2", ">i4"),
                     ("ir_operational_cal_ch_5_coefficient_3", ">i4"),
                     ("ir_test_cal_ch_5_coefficient_1", ">i4"),
                     ("ir_test_cal_ch_5_coefficient_2", ">i4"),
                     ("ir_test_cal_ch_5_coefficient_3", ">i4"),
                     ("zero_fill2", ">i4", (3, )),
                     # NAVIGATION
                     ("navigation_status_bit_field", ">u4"),
                     ("time_associated_with_tip_euler_angles", ">u4"),
                     ("tip_euler_angles", ">i2", (3, )),
                     ("spacecraft_altitude_above_reference_ellipsoid", ">u2"),
                     ("angular_relationships", ">i2", (153, )),
                     ("zero_fill3", ">i2", (3, )),
                     ("earth_location", ">i4", (102, )),
                     ("zero_fill4", ">i4", (2, )),
                     # HRPT MINOR FRAME TELEMETRY
                     ("frame_sync", ">u2", (6, )),
                     ("id", ">u2", (2, )),
                     ("time_code", ">u2", (4, )),
                     ("telemetry", ">u2", (10, )),
                     ("back_scan", ">u2", (30, )),
                     ("space_data", ">u2", (50, )),
                     ("sync_delta", ">u2"),
                     ("zero_fill5", ">i2"),
                     # AVHRR SENSOR DATA
                     ("sensor_data", ">u4", (682, )),
                     ("zero_fill6", ">i4", (2, )),
                     # DIGITAL B TELEMETRY
                     ("invalid_word_bit_flags1", ">u2"),
                     ("avhrr_digital_b_data", ">u2"),
                     ("zero_fill7", ">i4", (3, )),
                     # ANALOG HOUSEKEEPING DATA (TIP)
                     ("invalid_word_bit_flags2", ">u4"),
                     ("word_1_patch_temperature", ">u1", (22, )),
                     ("zero_fill8", ">i2", (3, )),
                     # CLOUDS FROM AVHRR (CLAVR)
                     ("reserved0", ">u4"),
                     ("reserved1", ">u4"),
                     ("reserved2", ">u2", (52, )),
                     # FILLER
                     ("zero_fill9", ">i4", (112, ))])


class KLMReader(GACReader):

    spacecraft_names = {4: 'noaa15',
                        2: 'noaa16',
                        6: 'noaa17',
                        7: 'noaa18',
                        8: 'noaa19',
                        12: 'metopa',
                        11: 'metopb',
                        }
    spacecrafts_orbital = {4: 'noaa 15',
                           2: 'noaa 16',
                           6: 'noaa 17',
                           7: 'noaa 18',
                           8: 'noaa 19',
                           12: 'metop 02',
                           11: 'metop 01',
                           }

    def read(self, filename):
        with open(filename) as fd_:
            self.head = np.fromfile(fd_, dtype=header, count=1)[0]
            fd_.seek(4608, 0)
            self.scans = np.fromfile(
                fd_, dtype=scanline, count=self.head["count_of_data_records"])

        self.spacecraft_id = self.head["noaa_spacecraft_identification_code"]
        self.spacecraft_name = self.spacecraft_names[self.spacecraft_id]

    def get_telemetry(self):
        prt_counts = np.mean(self.scans["telemetry"][:, 5:8], axis=1)

        # getting ICT counts

        ict_counts = np.zeros((len(self.scans), 3))
        ict_counts[:, 0] = np.mean(self.scans["back_scan"][:, 0::3], axis=1)
        ict_counts[:, 1] = np.mean(self.scans["back_scan"][:, 1::3], axis=1)
        ict_counts[:, 2] = np.mean(self.scans["back_scan"][:, 2::3], axis=1)

        # getting space counts

        space_counts = np.zeros((len(self.scans), 3))
        space_counts[:, 0] = np.mean(self.scans["space_data"][:, 2::5], axis=1)
        space_counts[:, 1] = np.mean(self.scans["space_data"][:, 3::5], axis=1)
        space_counts[:, 2] = np.mean(self.scans["space_data"][:, 4::5], axis=1)

        return prt_counts, ict_counts, space_counts

    def get_lonlat(self):
        arr_lat = self.scans["earth_location"][:, 0::2] / 1e4
        arr_lon = self.scans["earth_location"][:, 1::2] / 1e4

        self.lons, self.lats = gtp.Gac_Lat_Lon_Interpolator(arr_lon, arr_lat)
        return self.lons, self.lats

    def get_times(self):
        if self.utcs is None:
            year = self.scans["scan_line_year"]
            jday = self.scans["scan_line_day_of_year"]
            msec = self.scans["scan_line_utc_time_of_day"]

	    jday = np.where(np.logical_or(jday<1, jday>366),np.median(jday),jday)
	    if_wrong_jday = np.ediff1d(jday, to_begin=0)
	    jday = np.where(if_wrong_jday<0, max(jday), jday)

	    if_wrong_msec = np.where(msec<1)
	    if_wrong_msec = if_wrong_msec[0]
	    if len(if_wrong_msec) > 0:
		if if_wrong_msec[0] !=0:
		   msec = msec[0] + 0.5 * 1000.0 * (self.scans["scan_line_number"] - 1)
		else:
		   msec = np.median(msec - 0.5 * 1000.0 * (self.scans["scan_line_number"] - 1))	    
	    
	    if_wrong_msec = np.ediff1d(msec, to_begin=0)
            msec = np.where(np.logical_and(np.logical_or(if_wrong_msec<-1000, if_wrong_msec>1000),if_wrong_jday!=1), msec[0] + 0.5 * 1000.0 * (self.scans["scan_line_number"] - 1), msec)


            self.utcs = (((year - 1970).astype('datetime64[Y]')
                          + (jday - 1).astype('timedelta64[D]')).astype('datetime64[ms]')
                         + msec.astype('timedelta64[ms]'))
            self.times = self.utcs.astype(datetime.datetime)


	    # checking if year value is out of valid range
            if_wrong_year = np.where(np.logical_or(year<1978, year>2015))
            if_wrong_year = if_wrong_year[0]
            if len(if_wrong_year) > 0:
                # if the first scanline has valid time stamp
                if if_wrong_year[0] != 0:
                        year = year[0]
                        jday = jday[0]
                        msec = msec[0] + 0.5 * 1000.0 * (self.scans["scan_line_number"] - 1)
                        self.utcs = (((year - 1970).astype('datetime64[Y]')
                                      + (jday - 1).astype('timedelta64[D]')).astype('datetime64[ms]')
                                     + msec.astype('timedelta64[ms]'))
                        self.times = self.utcs.astype(datetime.datetime)
                # Otherwise use median time stamp
                else:
                        year = np.median(year)
                        jday = np.median(jday)
                        msec = np.median(msec - 0.5 * 1000.0 * (self.scans["scan_line_number"] - 1))
                        self.utcs = (((year - 1970).astype('datetime64[Y]')
                                      + (jday - 1).astype('timedelta64[D]')).astype('datetime64[ms]')
                                     + msec.astype('timedelta64[ms]'))
                        self.times = self.utcs.astype(datetime.datetime)

        return self.utcs

    def get_ch3_switch(self):
        """0 for 3b, 1 for 3a
        """
        return self.scans["scan_line_bit_field"][:] & 3

    def get_corrupt_mask(self):

        # corrupt scanlines

        mask = ((self.scans["quality_indicator_bit_field"] >> 31) |
                ((self.scans["quality_indicator_bit_field"] << 3) >> 31) |
                ((self.scans["quality_indicator_bit_field"] << 4) >> 31)) 
        
	number_of_scans = self.scans["telemetry"].shape[0]
        qual_flags = np.zeros((int(number_of_scans),7))
        qual_flags[:,0]=self.scans["scan_line_number"] 
        qual_flags[:,1]=(self.scans["quality_indicator_bit_field"] >> 31)
        qual_flags[:,2]=((self.scans["quality_indicator_bit_field"] << 3) >> 31)
        qual_flags[:,3]=((self.scans["quality_indicator_bit_field"] << 4) >> 31)
        qual_flags[:,4]=((self.scans["quality_indicator_bit_field"] << 24) >> 30)
        qual_flags[:,5]=((self.scans["quality_indicator_bit_field"] << 26) >> 30)
        qual_flags[:,6]=((self.scans["quality_indicator_bit_field"] << 28) >> 30)

	return mask.astype(bool), qual_flags


def main(filename, start_line, end_line):
    tic = datetime.datetime.now()
    reader = KLMReader()
    reader.read(filename)
    reader.get_lonlat()
    channels = reader.get_calibrated_channels()
    sat_azi, sat_zen, sun_azi, sun_zen, rel_azi = reader.get_angles()
  
    mask, qual_flags = reader.get_corrupt_mask()
    if (np.all(mask)):
        print "ERROR: All data is masked out. Stop processing"
        raise ValueError("All data is masked out.")
    
    gac_io.save_gac(reader.spacecraft_name,
                    reader.utcs,
                    reader.lats, reader.lons,
                    channels[:, :, 0], channels[:, :, 1],
                    channels[:, :, 2],
                    channels[:, :, 3],
                    channels[:, :, 4],
                    channels[:, :, 5],
                    sun_zen, sat_zen, sun_azi, sat_azi, rel_azi,
                    mask, qual_flags, start_line, end_line, reader.get_ch3_switch())
    LOG.info("pygac took: %s", str(datetime.datetime.now() - tic))

if __name__ == "__main__":
    import sys
    main(sys.argv[1], sys.argv[2], sys.argv[3])



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright (c) 2012, 2014 Abhay Devasthale

# Author(s):

#   Abhay Devasthale <abhay.devasthale@smhi.se>
#   Adam Dybbroe <adam.dybbroe@smhi.se>
#   Sara Hornquist <sara.hornquist@smhi.se>
#   Martin Raspaud <martin.raspaud@smhi.se>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


import h5py
import numpy as np
import time
import calendar
import datetime

import logging
LOG = logging.getLogger(__name__)

import ConfigParser
import os

try:
    CONFIG_FILE = os.environ['PYGAC_CONFIG_FILE']
except KeyError:
    LOG.exception('Environment variable PYGAC_CONFIG_FILE not set!')
    raise

if not os.path.exists(CONFIG_FILE) or not os.path.isfile(CONFIG_FILE):
    raise IOError(str(CONFIG_FILE) + " pointed to by the environment " +
                  "variable PYGAC_CONFIG_FILE is not a file or does not exist!")

conf = ConfigParser.ConfigParser()
try:
    conf.read(CONFIG_FILE)
except ConfigParser.NoSectionError:
    LOG.exception('Failed reading configuration file: ' + str(CONFIG_FILE))
    raise

options = {}
for option, value in conf.items('output', raw=True):
    options[option] = value

OUTDIR = options['output_dir']
OUTPUT_FILE_PREFIX = options['output_file_prefix']

SUNSATANGLES_DIR = os.environ.get('SM_SUNSATANGLES_DIR', OUTDIR)
AVHRR_DIR = os.environ.get('SM_AVHRR_DIR', OUTDIR)
QUAL_DIR = os.environ.get('SM_AVHRR_DIR', OUTDIR)
MISSING_DATA = -32001
MISSING_DATA_LATLON =  -999999


def save_gac(satellite_name,
             xutcs,
             lats, lons,
             ref1, ref2, ref3,
             bt3, bt4, bt5,
             sun_zen, sat_zen, sun_azi, sat_azi, rel_azi,
             mask, qual_flags, start_line, end_line, switch=None):

 
    start_line = int(start_line)
    end_line = int(end_line)

    bt3 = np.where(np.logical_or(bt3<170.0, bt3>350.0), MISSING_DATA, bt3-273.15) 
    bt4 = np.where(np.logical_or(bt4<170.0, bt4>350.0), MISSING_DATA, bt4-273.15) 
    bt5 = np.where(np.logical_or(bt5<170.0, bt5>350.0), MISSING_DATA, bt5-273.15) 

    lats = np.where(np.logical_or(lats<-90.00, lats>90.00), MISSING_DATA_LATLON, lats)
    lons = np.where(np.logical_or(lons<-180.00, lons>180.00), MISSING_DATA_LATLON, lons)

    sat_azi -= 180.0
    rel_azi = abs(rel_azi)
    rel_azi = 180.0 - rel_azi

    for array in [bt3, bt4, bt5]:
        array[array!=MISSING_DATA]=100*array[array!=MISSING_DATA]
        array[mask] = MISSING_DATA
    for array in [ref1, ref2, ref3,
                  sun_zen, sat_zen, sun_azi, sat_azi, rel_azi]:
        array *= 100        
        array[mask] = MISSING_DATA
    for array in [lats, lons]:
        array[array!=MISSING_DATA_LATLON]= 1000.0*array[array!=MISSING_DATA_LATLON]
        array[mask] = MISSING_DATA_LATLON

    for ref in [ref1, ref2, ref3]:
        ref[ref < 0] = MISSING_DATA

    if switch is not None:
        ref3[switch == 0] = MISSING_DATA
        bt3[switch == 1] = MISSING_DATA
        ref3[switch == 2] = MISSING_DATA
        bt3[switch == 2] = MISSING_DATA



    no_wrong_lat = np.where(lats!=MISSING_DATA_LATLON)	
    temp_start_line = min(no_wrong_lat[0]) 
    temp_end_line = max(no_wrong_lat[0])


    if temp_start_line>0 or temp_start_line>start_line:
       LOG.info('New start_line chosen (due to invalid lat/lon info) = ' + str(temp_start_line))
    if temp_end_line < lats.shape[0] or (end_line == 0 and temp_end_line < lats.shape[0]):
       LOG.info('New temporary end_line chosen (due to invalid lat/lon info) = ' + str(temp_end_line))
    if end_line>temp_end_line:
       end_line = temp_end_line
       LOG.info('New end_line chosen (due to invalid lat/lon info) = ' + str(temp_end_line))

  
    ref1 = ref1[temp_start_line:temp_end_line+1,:].copy()
    ref2 = ref2[temp_start_line:temp_end_line+1,:].copy()
    ref3 = ref3[temp_start_line:temp_end_line+1,:].copy()
    bt3 = bt3[temp_start_line:temp_end_line+1,:].copy()
    bt4 = bt4[temp_start_line:temp_end_line+1,:].copy()
    bt5 = bt5[temp_start_line:temp_end_line+1,:].copy()
    sun_zen = sun_zen[temp_start_line:temp_end_line+1,:].copy()
    sun_azi = sun_azi[temp_start_line:temp_end_line+1,:].copy()
    sat_zen = sat_zen[temp_start_line:temp_end_line+1,:].copy()
    sat_azi = sat_azi[temp_start_line:temp_end_line+1,:].copy()
    rel_azi = rel_azi[temp_start_line:temp_end_line+1,:].copy()
    lats = lats[temp_start_line:temp_end_line+1,:].copy()
    lons = lons[temp_start_line:temp_end_line+1,:].copy()
    qual_flags = qual_flags[temp_start_line:temp_end_line+1,:].copy()
    xutcs = xutcs[temp_start_line:temp_end_line+1].copy()


    # Reading time from the body of the gac file
    start = xutcs[start_line].astype(datetime.datetime)
    if end_line == 0:
        end = xutcs[-1].astype(datetime.datetime)
    else:
        end = xutcs[end_line].astype(datetime.datetime)

    startdate = start.strftime("%Y%m%d")
    starttime = start.strftime("%H%M%S%f")[:-5]
    enddate = end.strftime("%Y%m%d")
    endtime = end.strftime("%H%M%S%f")[:-5]
    jday = int(start.strftime("%j"))
   
    # Earth-Sun distance correction factor
    corr = 1.0 - 0.0334 * np.cos(2.0 * np.pi * (jday - 2) / 365.25)

   
    total_number_of_scan_lines = end_line - start_line + 1 
    last_scan_line_number = qual_flags[-1,0]


    if end_line>0:
       ref1 = ref1[start_line:end_line+1,:].copy()
       ref2 = ref2[start_line:end_line+1,:].copy()
       ref3 = ref3[start_line:end_line+1,:].copy()
       bt3 = bt3[start_line:end_line+1,:].copy()
       bt4 = bt4[start_line:end_line+1,:].copy()
       bt5 = bt5[start_line:end_line+1,:].copy()
       sun_zen = sun_zen[start_line:end_line+1,:].copy()
       sun_azi = sun_azi[start_line:end_line+1,:].copy()
       sat_zen = sat_zen[start_line:end_line+1,:].copy()
       sat_azi = sat_azi[start_line:end_line+1,:].copy()
       rel_azi = rel_azi[start_line:end_line+1,:].copy()
       lats = lats[start_line:end_line+1,:].copy()
       lons = lons[start_line:end_line+1,:].copy()
       qual_flags = qual_flags[start_line:end_line+1,:].copy()
        

    avhrrGAC_io(satellite_name, startdate, enddate, starttime, endtime,
                lats, lons, ref1, ref2, ref3, bt3, bt4, bt5,
                sun_zen, sat_zen, sun_azi, sat_azi, rel_azi, qual_flags, 
		start_line, end_line, total_number_of_scan_lines,
                last_scan_line_number, corr)


def avhrrGAC_io(satellite_name, startdate, enddate, starttime, endtime,
                arrLat_full, arrLon_full, ref1, ref2, ref3, bt3, bt4, bt5,
                arrSZA, arrSTZ, arrSAA, arrSTA, arrRAA, qual_flags, 
		start_line, end_line, total_number_of_scan_lines,
                last_scan_line_number, corr):
    import os

    # Calculate start and end time in sec1970
    t_obj = time.strptime(startdate + starttime[0:6], "%Y%m%d%H%M%S")
    starttime_sec1970 = calendar.timegm(t_obj)
    t_obj = time.strptime(enddate + endtime[0:6], "%Y%m%d%H%M%S")
    endtime_sec1970 = calendar.timegm(t_obj)

    LOG.info('Output file prefix = ' + str(OUTPUT_FILE_PREFIX))
    LOG.info('AVHRR data will be written to ' + str(AVHRR_DIR))
    ofn = os.path.join(AVHRR_DIR, (OUTPUT_FILE_PREFIX + '_avhrr_' +
                                   satellite_name + '_99999_' +
                                   startdate + 'T' + starttime + 'Z_' +
                                   enddate + 'T' + endtime + 'Z.h5'))
    
    LOG.info('Filename: ' + str(os.path.basename(ofn)))

    fout = h5py.File(ofn, "w")

    dset1 = fout.create_dataset("/image1/data", dtype='int16', data=ref1)
    dset2 = fout.create_dataset("/image2/data", dtype='int16', data=ref2)
    dset3 = fout.create_dataset("/image3/data", dtype='int16', data=bt3)
    dset4 = fout.create_dataset("/image4/data", dtype='int16', data=bt4)
    dset5 = fout.create_dataset("/image5/data", dtype='int16', data=bt5)
    dset6 = fout.create_dataset("/image6/data", dtype='int16', data=ref3)
    dset7 = fout.create_dataset("/where/lat/data", dtype='int32',
                                data=arrLat_full)
    dset8 = fout.create_dataset("/where/lon/data", dtype='int32',
                                data=arrLon_full)

    channellist = []
    channellist.append("channel1")
    channellist.append("channel2")
    channellist.append("channel3b")
    channellist.append("channel4")
    channellist.append("channel5")
    channellist.append("channel3a")
    dset10 = fout.create_dataset("/how/channel_list",
                                 data=channellist)

    # Attributes directly on highest level groups
    g1 = fout.require_group("/image1")
    g2 = fout.require_group("/image2")
    g3 = fout.require_group("/image3")
    g4 = fout.require_group("/image4")
    g5 = fout.require_group("/image5")
    g6 = fout.require_group("/image6")
    g7 = fout.require_group("/where")

    g1.attrs["channel"] = np.string_("1")
    g1.attrs["description"] = np.string_("AVHRR ch1")
    g2.attrs["channel"] = np.string_("2")
    g2.attrs["description"] = np.string_("AVHRR ch2")
    g3.attrs["channel"] = np.string_("3b")
    g3.attrs["description"] = np.string_("AVHRR ch3b")
    g4.attrs["channel"] = np.string_("4")
    g4.attrs["description"] = np.string_("AVHRR ch4")
    g5.attrs["channel"] = np.string_("5")
    g5.attrs["description"] = np.string_("AVHRR ch5")
    g6.attrs["channel"] = np.string_("3a")
    g6.attrs["description"] = np.string_("AVHRR ch3a")
    g7.attrs["num_of_pixels"] = np.int32(arrSZA.shape[1])
    g7.attrs["num_of_lines"] = np.int32(arrSZA.shape[0])
    g7.attrs["xscale"] = np.float32(0.0)
    g7.attrs["yscale"] = np.float32(0.0)
    g7.attrs["start_line"] = start_line
    g7.attrs["end_line"] = end_line

    # Attributes in the 'what' groups
    g1 = fout.create_group("/image1/what")
    g2 = fout.create_group("/image2/what")
    g3 = fout.create_group("/image3/what")
    g4 = fout.create_group("/image4/what")
    g5 = fout.create_group("/image5/what")
    g6 = fout.create_group("/image6/what")
    g7 = fout.create_group("/where/lat/what")
    g8 = fout.create_group("/where/lon/what")
    g9 = fout.create_group("/what")

    g1.attrs["product"] = np.string_("SATCH")
    g1.attrs["quantity"] = np.string_("REFL")
    g1.attrs["dataset_name"] = np.string_('Channel 1 reflectance')
    g1.attrs["units"] = np.string_('%')
    g1.attrs["gain"] = np.float32(0.01)
    g1.attrs["offset"] = np.float32(0.0)
    g1.attrs["missingdata"] = np.int32(MISSING_DATA)
    g1.attrs["nodata"] = np.int32(MISSING_DATA)
    g1.attrs["starttime"] = np.string_(starttime[0:6])
    g1.attrs["endtime"] = np.string_(endtime[0:6])
    g1.attrs["startdate"] = np.string_(startdate)
    g1.attrs["enddate"] = np.string_(enddate)

    g2.attrs["product"] = np.string_("SATCH")
    g2.attrs["quantity"] = np.string_("REFL")
    g2.attrs["dataset_name"] = np.string_('Channel 2 reflectance')
    g2.attrs["units"] = np.string_('%')
    g2.attrs["gain"] = np.float32(0.01)
    g2.attrs["offset"] = np.float32(0.0)
    g2.attrs["missingdata"] = np.int32(MISSING_DATA)
    g2.attrs["nodata"] = np.int32(MISSING_DATA)
    g2.attrs["starttime"] = np.string_(starttime[0:6])
    g2.attrs["endtime"] = np.string_(endtime[0:6])
    g2.attrs["startdate"] = np.string_(startdate)
    g2.attrs["enddate"] = np.string_(enddate)

    g6.attrs["product"] = np.string_("SATCH")
    g6.attrs["quantity"] = np.string_("REFL")
    g6.attrs["dataset_name"] = np.string_('Channel 3a reflectance')
    g6.attrs["units"] = np.string_('%')
    g6.attrs["gain"] = np.float32(0.01)
    g6.attrs["offset"] = np.float32(0.0)
    g6.attrs["missingdata"] = np.int32(MISSING_DATA)
    g6.attrs["nodata"] = np.int32(MISSING_DATA)
    g6.attrs["starttime"] = np.string_(starttime[0:6])
    g6.attrs["endtime"] = np.string_(endtime[0:6])
    g6.attrs["startdate"] = np.string_(startdate)
    g6.attrs["enddate"] = np.string_(enddate)

    g3.attrs["product"] = np.string_("SATCH")
    g3.attrs["quantity"] = np.string_("TB")
    g3.attrs["dataset_name"] = np.string_('Channel 3b brightness temperature')
    g3.attrs["units"] = np.string_('K')
    g3.attrs["gain"] = np.float32(0.01)
    g3.attrs["offset"] = np.float32(273.15)
    g3.attrs["missingdata"] = np.int32(MISSING_DATA)
    g3.attrs["nodata"] = np.int32(MISSING_DATA)
    g3.attrs["starttime"] = np.string_(starttime[0:6])
    g3.attrs["endtime"] = np.string_(endtime[0:6])
    g3.attrs["startdate"] = np.string_(startdate)
    g3.attrs["enddate"] = np.string_(enddate)

    g4.attrs["product"] = np.string_("SATCH")
    g4.attrs["quantity"] = np.string_("TB")
    g4.attrs["dataset_name"] = np.string_('Channel 4 brightness temperature')
    g4.attrs["units"] = np.string_('K')
    g4.attrs["gain"] = np.float32(0.01)
    g4.attrs["offset"] = np.float32(273.15)
    g4.attrs["missingdata"] = np.int32(MISSING_DATA)
    g4.attrs["nodata"] = np.int32(MISSING_DATA)
    g4.attrs["starttime"] = np.string_(starttime[0:6])
    g4.attrs["endtime"] = np.string_(endtime[0:6])
    g4.attrs["startdate"] = np.string_(startdate)
    g4.attrs["enddate"] = np.string_(enddate)

    g5.attrs["product"] = np.string_("SATCH")
    g5.attrs["quantity"] = np.string_("TB")
    g5.attrs["dataset_name"] = np.string_('Channel 5 brightness temperature')
    g5.attrs["units"] = np.string_('K')
    g5.attrs["gain"] = np.float32(0.01)
    g5.attrs["offset"] = np.float32(273.15)
    g5.attrs["missingdata"] = np.int32(MISSING_DATA)
    g5.attrs["nodata"] = np.int32(MISSING_DATA)
    g5.attrs["starttime"] = np.string_(starttime[0:6])
    g5.attrs["endtime"] = np.string_(endtime[0:6])
    g5.attrs["startdate"] = np.string_(startdate)
    g5.attrs["enddate"] = np.string_(enddate)

    g7.attrs["dataset_name"] = np.string_('Latitude')
    g7.attrs["units"] = np.string_('Deg')
    g7.attrs["gain"] = np.float32(0.0010)
    g7.attrs["offset"] = np.float32(0.0)
    g7.attrs["missingdata"] = np.int32(MISSING_DATA_LATLON)
    g7.attrs["nodata"] = np.int32(MISSING_DATA_LATLON)
    g7.attrs["starttime"] = np.string_(starttime[0:6])
    g7.attrs["endtime"] = np.string_(endtime[0:6])
    g7.attrs["startdate"] = np.string_(startdate)
    g7.attrs["enddate"] = np.string_(enddate)

    g8.attrs["dataset_name"] = np.string_('Longitude')
    g8.attrs["units"] = np.string_('Deg')
    g8.attrs["gain"] = np.float32(0.0010)
    g8.attrs["offset"] = np.float32(0.0)
    g8.attrs["missingdata"] = np.int32(MISSING_DATA_LATLON)
    g8.attrs["nodata"] = np.int32(MISSING_DATA_LATLON)
    g8.attrs["starttime"] = np.string_(starttime[0:6])
    g8.attrs["endtime"] = np.string_(endtime[0:6])
    g8.attrs["startdate"] = np.string_(startdate)
    g8.attrs["enddate"] = np.string_(enddate)

    g9.attrs["object"] = np.string_("SATP")
    g9.attrs["sets"] = np.int32(len(channellist))
    g9.attrs["version"] = np.string_("H5rad ?.?")
    g9.attrs["date"] = np.string_(startdate)
    g9.attrs["time"] = np.string_(starttime[0:6])

    # Attributes in the 'how' groups
    g1 = fout.create_group("/image1/how")
    g2 = fout.create_group("/image2/how")
    g3 = fout.create_group("/image3/how")
    g4 = fout.create_group("/image4/how")
    g5 = fout.create_group("/image5/how")
    g6 = fout.create_group("/image6/how")
    g10 = fout.require_group("/how")

    # SHq: Is the sun_earth_distance correction applied?
    g1.attrs["sun_earth_distance_correction_applied"] = np.string_("TRUE")
    g1.attrs["sun_earth_distance_correction_factor"] = corr 
    g2.attrs["sun_earth_distance_correction_applied"] = np.string_("TRUE")
    g2.attrs["sun_earth_distance_correction_factor"] = corr
    # No attributes on 'how' for image3,4,5
    g6.attrs["sun_earth_distance_correction_applied"] = np.string_("TRUE")
    g6.attrs["sun_earth_distance_correction_factor"] = corr

    # We do not know much about how; mostly use no-data
    g10.attrs["yaw_error"] = 0.0
    g10.attrs["roll_error"] = 0.0
    g10.attrs["pitch_error"] = 0.0
    g10.attrs["startepochs"] = starttime_sec1970
    g10.attrs["endepochs"] = endtime_sec1970
    g10.attrs["platform"] = np.string_(satellite_name)
    g10.attrs["instrument"] = np.string_("avhrr")
    g10.attrs["orbit_number"] = np.int32(99999)
    g10.attrs["software"] = np.string_("pyGAC")
    g10.attrs["version"] = np.string_("1.0")

    fout.close()

    LOG.info('Sun and Satellite viewing angles will be ' +
             'written to ' + str(SUNSATANGLES_DIR))
    ofn = os.path.join(SUNSATANGLES_DIR,
                       (OUTPUT_FILE_PREFIX + '_sunsatangles_' +
                        satellite_name + '_99999_' + startdate +
                        'T' + starttime + 'Z_' +
                        enddate + 'T' + endtime + 'Z.h5')) 

    LOG.info('Filename: ' + str(os.path.basename(ofn)))
    fout = h5py.File(ofn, "w")

    dset1 = fout.create_dataset("/image1/data", dtype='int16', data=arrSZA)
    dset2 = fout.create_dataset("/image2/data", dtype='int16', data=arrSTZ)
    dset3 = fout.create_dataset("/image3/data", dtype='int16', data=arrRAA)
    dset4 = fout.create_dataset("/image4/data", dtype='int16', data=arrSAA)
    dset5 = fout.create_dataset("/image5/data", dtype='int16', data=arrSTA)
    dset6 = fout.create_dataset("/where/lat/data", dtype='int32',
                                data=arrLat_full)
    dset7 = fout.create_dataset("/where/lon/data", dtype='int32',
                                data=arrLon_full)

    # Attributes directly on highest level groups
    g1 = fout.require_group("/image1")
    g2 = fout.require_group("/image2")
    g3 = fout.require_group("/image3")
    g4 = fout.require_group("/image4")
    g5 = fout.require_group("/image5")
    g6 = fout.require_group("/where")

    g1.attrs["description"] = np.string_('Solar zenith angle')
    g2.attrs["description"] = np.string_('Satellite zenith angle')
    g3.attrs["description"] = np.string_(
        'Relative satellite-sun azimuth angle')
    g4.attrs["description"] = np.string_('Solar azimuth angle')
    g5.attrs["description"] = np.string_('Satellite azimuth angle')
    g6.attrs["num_of_pixels"] = np.int32(arrSZA.shape[1])
    g6.attrs["num_of_lines"] = np.int32(arrSZA.shape[0])
    g6.attrs["xscale"] = np.float32(0.0)
    g6.attrs["yscale"] = np.float32(0.0)
    g6.attrs["start_line"] = start_line
    g6.attrs["end_line"] = end_line

    # Attributes in the 'what' groups + 'how'
    g1 = fout.create_group("/image1/what")
    g2 = fout.create_group("/image2/what")
    g3 = fout.create_group("/image3/what")
    g4 = fout.create_group("/image4/what")
    g5 = fout.create_group("/image5/what")
    g6 = fout.create_group("/where/lat/what")
    g7 = fout.create_group("/where/lon/what")
    g8 = fout.create_group("/what")
    g9 = fout.create_group("/how")

    g1.attrs["product"] = np.string_("SUNZ")
    g1.attrs["quantity"] = np.string_("DEG")
    g1.attrs["dataset_name"] = np.string_('Solar zenith angle')
    g1.attrs["units"] = np.string_('Deg')
    g1.attrs["gain"] = np.float32(0.01)
    g1.attrs["offset"] = np.float32(0.0)
    g1.attrs["missingdata"] = np.int32(MISSING_DATA)
    g1.attrs["nodata"] = np.int32(MISSING_DATA)
    g1.attrs["starttime"] = np.string_(starttime[0:6])
    g1.attrs["endtime"] = np.string_(endtime[0:6])
    g1.attrs["startdate"] = np.string_(startdate)
    g1.attrs["enddate"] = np.string_(enddate)

    g2.attrs["product"] = np.string_("SATZ")
    g2.attrs["quantity"] = np.string_("DEG")
    g2.attrs["dataset_name"] = np.string_('Satellite zenith angle')
    g2.attrs["units"] = np.string_('Deg')
    g2.attrs["gain"] = np.float32(0.01)
    g2.attrs["offset"] = np.float32(0.0)
    g2.attrs["missingdata"] = np.int32(MISSING_DATA)
    g2.attrs["nodata"] = np.int32(MISSING_DATA)
    g2.attrs["starttime"] = np.string_(starttime[0:6])
    g2.attrs["endtime"] = np.string_(endtime[0:6])
    g2.attrs["startdate"] = np.string_(startdate)
    g2.attrs["enddate"] = np.string_(enddate)

    g3.attrs["product"] = np.string_("SSAZD")
    g3.attrs["quantity"] = np.string_("DEG")
    g3.attrs["dataset_name"] = np.string_( \
        'Relative satellite-sun azimuth angle')
    g3.attrs["units"] = np.string_('Deg')
    g3.attrs["gain"] = np.float32(0.01)
    g3.attrs["offset"] = np.float32(0.0)
    g3.attrs["missingdata"] = np.int32(MISSING_DATA)
    g3.attrs["nodata"] = np.int32(MISSING_DATA)
    g3.attrs["starttime"] = np.string_(starttime[0:6])
    g3.attrs["endtime"] = np.string_(endtime[0:6])
    g3.attrs["startdate"] = np.string_(startdate)
    g3.attrs["enddate"] = np.string_(enddate)

    g4.attrs["product"] = np.string_("SUNA")
    g4.attrs["quantity"] = np.string_("DEG")
    g4.attrs["dataset_name"] = np.string_('Solar azimuth angle')
    g4.attrs["units"] = np.string_('Deg')
    g4.attrs["gain"] = np.float32(0.01)
    g4.attrs["offset"] = np.float32(180.0)
    g4.attrs["missingdata"] = np.int32(MISSING_DATA)
    g4.attrs["nodata"] = np.int32(MISSING_DATA)
    g4.attrs["starttime"] = np.string_(starttime[0:6])
    g4.attrs["endtime"] = np.string_(endtime[0:6])
    g4.attrs["startdate"] = np.string_(startdate)
    g4.attrs["enddate"] = np.string_(enddate)

    g5.attrs["product"] = np.string_("SATA")
    g5.attrs["quantity"] = np.string_("DEG")
    g5.attrs["dataset_name"] = np.string_('Satellite azimuth angle')
    g5.attrs["units"] = np.string_('Deg')
    g5.attrs["gain"] = np.float32(0.01)
    g5.attrs["offset"] = np.float32(180.0)
    g5.attrs["missingdata"] = np.int32(MISSING_DATA)
    g5.attrs["nodata"] = np.int32(MISSING_DATA)
    g5.attrs["starttime"] = np.string_(starttime[0:6])
    g5.attrs["endtime"] = np.string_(endtime[0:6])
    g5.attrs["startdate"] = np.string_(startdate)
    g5.attrs["enddate"] = np.string_(enddate)

    g6.attrs["dataset_name"] = np.string_('Latitude')
    g6.attrs["units"] = np.string_('Deg')
    g6.attrs["gain"] = np.float32(0.0010)
    g6.attrs["offset"] = np.float32(0.0)
    g6.attrs["missingdata"] = np.int32(MISSING_DATA_LATLON)
    g6.attrs["nodata"] = np.int32(MISSING_DATA_LATLON)
    g6.attrs["starttime"] = np.string_(starttime[0:6])
    g6.attrs["endtime"] = np.string_(endtime[0:6])
    g6.attrs["startdate"] = np.string_(startdate)
    g6.attrs["enddate"] = np.string_(enddate)

    g7.attrs["dataset_name"] = np.string_('Longitude')
    g7.attrs["units"] = np.string_('Deg')
    g7.attrs["gain"] = np.float32(0.0010)
    g7.attrs["offset"] = np.float32(0.0)
    g7.attrs["missingdata"] = np.int32(MISSING_DATA_LATLON)
    g7.attrs["nodata"] = np.int32(MISSING_DATA_LATLON)
    g7.attrs["starttime"] = np.string_(starttime[0:6])
    g7.attrs["endtime"] = np.string_(endtime[0:6])
    g7.attrs["startdate"] = np.string_(startdate)
    g7.attrs["enddate"] = np.string_(enddate)

    g8.attrs["object"] = np.string_("SATP")
    g8.attrs["sets"] = np.int32(5)
    g8.attrs["version"] = np.string_("H5rad ?.?")
    g8.attrs["date"] = np.string_(startdate)
    g8.attrs["time"] = np.string_(starttime[0:6])

    # We do not know much about how; mostly use no-data
    g9.attrs["yaw_error"] = 0.0
    g9.attrs["roll_error"] = 0.0
    g9.attrs["pitch_error"] = 0.0
    g9.attrs["startepochs"] = starttime_sec1970
    g9.attrs["endepochs"] = endtime_sec1970
    g9.attrs["platform"] = np.string_(satellite_name)
    g9.attrs["instrument"] = np.string_("avhrr")
    g9.attrs["orbit_number"] = np.int32(99999)
    g9.attrs["software"] = np.string_("pyGAC")
    g9.attrs["version"] = np.string_("1.0")

    fout.close()


    LOG.info('Quality flags will be ' +
             'written to ' + str(QUAL_DIR))
    ofn = os.path.join(QUAL_DIR,
                       (OUTPUT_FILE_PREFIX + '_qualflags_' +
                        satellite_name + '_99999_' + startdate +
                        'T' + starttime + 'Z_' +
                        enddate + 'T' + endtime + 'Z.h5')) 

    LOG.info('Filename: ' + str(os.path.basename(ofn)))
    fout = h5py.File(ofn, "w")

    dset1 = fout.create_dataset("/qual_flags/data", dtype='int16', data=qual_flags)

    g1 = fout.require_group("/qual_flags")

    g1.attrs["product"] = np.string_("QFLAG")
    g1.attrs["quantity"] = np.string_("INT")
    g1.attrs["dataset_name"] = np.string_('Scanline quality flags')
    g1.attrs["units"] = np.string_('None')
    g1.attrs["gain"] = np.int32(1)
    g1.attrs["offset"] = np.int32(0)
    g1.attrs["missingdata"] = np.int32(MISSING_DATA)
    g1.attrs["nodata"] = np.int32(MISSING_DATA)
    g1.attrs["starttime"] = np.string_(starttime[0:6])
    g1.attrs["endtime"] = np.string_(endtime[0:6])
    g1.attrs["startdate"] = np.string_(startdate)
    g1.attrs["enddate"] = np.string_(enddate)
    g1.attrs["total_number_of_data_records"] = total_number_of_scan_lines
    g1.attrs["last_scan_line_number"] = last_scan_line_number 

    fout.close()




#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright (c) 2010-2012.

# Author(s):
 
#   Adam Dybbroe <adam.dybbroe@smhise>
#   Martin Raspaud <martin.raspaud@smhi.se>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Interpolation of geographical tiepoints.
"""

import numpy as np
from numpy import arccos, sign, rad2deg, sqrt, arcsin
from scipy.interpolate import RectBivariateSpline, splrep, splev


EARTH_RADIUS = 6370997.0


def metop20kmto1km(lons20km, lats20km):
    """Getting 1km geolocation for metop avhrr from 20km tiepoints.
    """
    cols20km = np.array([0] + range(4, 2048, 20) + [2047])
    cols1km = np.arange(2048)
    lines = lons20km.shape[0]
    rows20km = np.arange(lines)
    rows1km = np.arange(lines)

    along_track_order = 1
    cross_track_order = 3

    satint = SatelliteInterpolator((lons20km, lats20km),
                                   (rows20km, cols20km),
                                   (rows1km, cols1km),
                                   along_track_order,
                                   cross_track_order)
    return satint.interpolate()

def modis5kmto1km(lons5km, lats5km):
    """Getting 1km geolocation for modis from 5km tiepoints.
    """
    cols5km = np.arange(2, 1354, 5)
    cols1km = np.arange(1354)
    lines = lons5km.shape[0] * 5
    rows5km = np.arange(2, lines, 5)
    rows1km = np.arange(lines)

    along_track_order = 1
    cross_track_order = 3

    satint = SatelliteInterpolator((lons5km, lats5km),
                                   (rows5km, cols5km),
                                   (rows1km, cols1km),
                                   along_track_order,
                                   cross_track_order,
                                   chunk_size=10)
    satint.fill_borders("y", "x")
    lons1km, lats1km = satint.interpolate()
    return lons1km, lats1km

def modis1kmto500m(lons1km, lats1km):
    """Getting 500m geolocation for modis from 1km tiepoints.
    """
    cols1km = np.arange(0, 2708, 2)
    cols500m = np.arange(2708)
    lines = lons1km.shape[0] * 2
    rows1km = np.arange(0.5, lines, 2)
    rows500m = np.arange(lines)

    along_track_order = 1
    cross_track_order = 3
    
    satint = SatelliteInterpolator((lons1km, lats1km),
                                   (rows1km, cols1km),
                                   (rows500m, cols500m),
                                   along_track_order,
                                   cross_track_order,
                                   chunk_size=20)
    satint.fill_borders("y", "x")
    lons500m, lats500m = satint.interpolate()
    return lons500m, lats500m

def modis1kmto250m(lons1km, lats1km):
    """Getting 250m geolocation for modis from 1km tiepoints.
    """
    cols1km = np.arange(0, 5416, 4)
    cols250m = np.arange(5416)
    lines = lons1km.shape[0] * 4
    rows1km = np.arange(1.5, lines, 4)
    rows250m = np.arange(lines)

    along_track_order = 1
    cross_track_order = 3
    
    satint = SatelliteInterpolator((lons1km, lats1km),
                                   (rows1km, cols1km),
                                   (rows250m, cols250m),
                                   along_track_order,
                                   cross_track_order,
                                   chunk_size=40)
    satint.fill_borders("y", "x")
    lons250m, lats250m = satint.interpolate()
    return lons250m, lats250m


def Gac_Lat_Lon_Interpolator(lons_subset,lats_subset):
    """ interpolates lat-lon values in the AVHRR GAC data from every eigth pixel to all pixels 
        Each GAC row has total 409 pixels. But lat-lon values are provided for every eigth pixel starting from pixel 5 and ending at pixel 405 """
    
    #cols_subset = np.arange(0, 404, 8)
    #cols_full = np.arange(405)
    cols_subset = np.arange(4, 405, 8)
    cols_full = np.arange(409)
    lines = lats_subset.shape[0]
    rows_subset = np.arange(lines)
    rows_full = np.arange(lines)

    along_track_order = 1
    cross_track_order = 3

    satint = SatelliteInterpolator((lons_subset, lats_subset),
                                   (rows_subset, cols_subset),
                                   (rows_full, cols_full),
                                   along_track_order,
                                   cross_track_order)

    return satint.interpolate()
    

# NOTE: extrapolate on a sphere ?
def _linear_extrapolate(pos, data, xev):
    """

    >>> import numpy as np
    >>> pos = np.array([1, 2])
    >>> data = np.arange(10).reshape((2, 5), order="F")
    >>> xev = 5
    >>> _linear_extrapolate(pos, data, xev)
    array([  4.,   6.,   8.,  10.,  12.])
    >>> xev = 0
    >>> _linear_extrapolate(pos, data, xev)
    array([-1.,  1.,  3.,  5.,  7.])
    """
    if len(data) != 2 or len(pos) != 2:
        raise ValueError("len(pos) and the number of lines of data"
                         " must be 2.")

    return data[1] + ((xev - pos[1]) / (1.0 * (pos[0] - pos[1])) *
                      (data[0] - data[1]))

class SatelliteInterpolator(object):
    """
    Handles interpolation of geolocation data from a grid of tie points.  It is
    preferable to have tie-points out till the edges if the tiepoint grid, but
    a method is provided to extrapolate linearly the tiepoints to the borders
    of the grid.

    Uses numpy, scipy, and pyresample

    The constructor takes in the tiepointed lon/lat data as *lon_lat_data*, the
    *tiepoint_grid* and the desired *final_grid*. As optional arguments, one
    can provide *kx_* and *ky_* as interpolation orders (in x and y directions
    respectively), and the *chunksize* if the data has to be handled by pieces
    along the y axis (this affects how the extrapolator behaves). If
    *chunksize* is set, don't forget to adjust the interpolation orders
    accordingly: the interpolation is indeed done globaly (not chunkwise).
    """

    def __init__(self, lon_lat_data, tiepoint_grid, final_grid,
                 kx_=1, ky_=1, chunk_size=0):
        self.row_indices = tiepoint_grid[0]
        self.col_indices = tiepoint_grid[1]
        self.hrow_indices = final_grid[0]
        self.hcol_indices = final_grid[1]
        self.chunk_size = chunk_size
        self.lon_tiepoint = None
        self.lat_tiepoint = None
        self.longitude = None
        self.latitude = None

        try:
            # Maybe it's a pyresample object ?
            self.set_tiepoints(lon_lat_data.lons, lon_lat_data.lats)
            xyz = lon_lat_data.get_cartesian_coords()
            self.x__ = xyz[:, :, 0]
            self.y__ = xyz[:, :, 1]
            self.z__ = xyz[:, :, 2]

        except AttributeError:
            self.set_tiepoints(lon_lat_data[0], lon_lat_data[1])
            lons_rad = np.radians(self.lon_tiepoint)
            lats_rad = np.radians(self.lat_tiepoint)
            self.x__ = EARTH_RADIUS * np.cos(lats_rad) * np.cos(lons_rad)
            self.y__ = EARTH_RADIUS * np.cos(lats_rad) * np.sin(lons_rad)
            self.z__ = EARTH_RADIUS * np.sin(lats_rad)

        self.newx = None
        self.newy = None
        self.newz = None

        self.kx_, self.ky_ = kx_, ky_
        
    def set_tiepoints(self, lon, lat):
        """Defines the lon,lat tie points.
        """
        self.lon_tiepoint = lon
        self.lat_tiepoint = lat

    def fill_borders(self, *args):
        """Extrapolate tiepoint lons and lats to fill in the border of the
        chunks.

        >>> import numpy as np
        >>> lons = np.arange(20).reshape((4, 5), order="F")
        >>> lats = np.arange(20).reshape((4, 5), order="C")
        >>> lines = np.array([2, 7, 12, 17])
        >>> cols = np.array([2, 7, 12, 17, 22])
        >>> hlines = np.arange(20)
        >>> hcols = np.arange(24)
        >>> satint = SatelliteInterpolator((lons, lats), (lines, cols), (hlines, hcols), chunk_size=10)
        >>> satint.fill_borders('x', 'y')
        >>> satint.x__
        array([[ 6384905.78040055,  6381081.08333225,  6371519.34066148,
                 6328950.00792935,  6253610.69157758,  6145946.19489936,
                 6124413.29556372],
               [ 6377591.95940176,  6370997.        ,  6354509.6014956 ,
                 6305151.62592155,  6223234.99818839,  6109277.14889072,
                 6086485.57903118],
               [ 6359307.40690478,  6345786.79166939,  6311985.2535809 ,
                 6245655.67090206,  6147295.76471541,  6017604.5338691 ,
                 5991666.28769983],
               [ 6351993.58590599,  6335702.70833714,  6294975.51441502,
                 6221857.28889426,  6116920.07132621,  5980935.48786045,
                 5953738.5711673 ],
               [ 6338032.26190294,  6320348.4990906 ,  6276139.09205974,
                 6199670.56624433,  6091551.90273768,  5952590.38414781,
                 5924798.08042984],
               [ 6290665.5946295 ,  6270385.16249031,  6219684.08214232,
                 6137100.75832981,  6023313.2794414 ,  5879194.72399075,
                 5850371.01290062],
               [ 6172248.92644589,  6145476.82098957,  6078546.55734877,
                 5980676.23854351,  5852716.72120069,  5695705.57359808,
                 5664303.34407756],
               [ 6124882.25917245,  6095513.48438928,  6022091.54743135,
                 5918106.430629  ,  5784478.09790441,  5622309.91344102,
                 5589876.27654834]])
        >>> satint.row_indices
        array([ 0,  2,  7,  9, 10, 12, 17, 19])
        >>> satint.col_indices
        array([ 0,  2,  7, 12, 17, 22, 23])
        """
        
        to_run = []
        cases = {"y": self._fill_row_borders,
                 "x": self._fill_col_borders}
        for dim in args:
            try:
                to_run.append(cases[dim])
            except KeyError:
                raise NameError("Unrecognized dimension: "+str(dim))

        for fun in to_run:
            fun()


    def _extrapolate_cols(self, data, first=True, last=True):
        """Extrapolate the column of data, to get the first and last together
        with the data.

        >>> import numpy as np
        >>> lons = np.arange(10).reshape((2, 5), order="F")
        >>> lats = np.arange(10).reshape((2, 5), order="C")
        >>> lines = np.array([2, 7])
        >>> cols = np.array([2, 7, 12, 17, 22])
        >>> hlines = np.arange(10)
        >>> hcols = np.arange(24)
        >>> satint = SatelliteInterpolator((lons, lats), (lines, cols), (hlines, hcols))
        >>> satint._extrapolate_cols(satint.x__)
        array([[ 6372937.31273379,  6370997.        ,  6366146.21816553,
                 6351605.98629588,  6327412.61244969,  6293626.50067273,
                 6286869.27831734],
               [ 6353136.46335726,  6345786.79166939,  6327412.61244969,
                 6299445.69529922,  6261968.60390423,  6215087.60607344,
                 6205711.40650728]])
        """

        if first:
            pos = self.col_indices[:2]
            first_column = _linear_extrapolate(pos,
                                               (data[:, 0], data[:, 1]),
                                               self.hcol_indices[0])
        if last:
            pos = self.col_indices[-2:]
            last_column = _linear_extrapolate(pos,
                                              (data[:, -2], data[:, -1]),
                                              self.hcol_indices[-1])

        if first and last:
            return np.hstack((np.expand_dims(first_column, 1),
                              data,
                              np.expand_dims(last_column, 1))) 
        elif first:
            return np.hstack((np.expand_dims(first_column, 1),
                              data))
        elif last:
            return np.hstack((data,
                              np.expand_dims(last_column, 1)))
        else:
            return data


    def _fill_col_borders(self):
        """Add the first and last column to the data by extrapolation.

        >>> import numpy as np
        >>> lons = np.arange(10).reshape((2, 5), order="F")
        >>> lats = np.arange(10).reshape((2, 5), order="C")
        >>> lines = np.array([2, 7])
        >>> cols = np.array([2, 7, 12, 17, 22])
        >>> hlines = np.arange(10)
        >>> hcols = np.arange(24)
        >>> satint = SatelliteInterpolator((lons, lats), (lines, cols), (hlines, hcols))
        >>> satint._fill_col_borders()
        >>> satint.x__
        array([[ 6372937.31273379,  6370997.        ,  6366146.21816553,
                 6351605.98629588,  6327412.61244969,  6293626.50067273,
                 6286869.27831734],
               [ 6353136.46335726,  6345786.79166939,  6327412.61244969,
                 6299445.69529922,  6261968.60390423,  6215087.60607344,
                 6205711.40650728]])
        >>> satint.col_indices
        array([ 0,  2,  7, 12, 17, 22, 23])
        """
        
        first = True
        last = True
        if self.col_indices[0] == self.hcol_indices[0]:
            first = False
        if self.col_indices[-1] == self.hcol_indices[-1]:
            last = False        

        self.x__ = self._extrapolate_cols(self.x__, first, last)
        self.y__ = self._extrapolate_cols(self.y__, first, last)
        self.z__ = self._extrapolate_cols(self.z__, first, last)
  
        if first and last:
            self.col_indices = np.concatenate((np.array([self.hcol_indices[0]]),
                                               self.col_indices,
                                               np.array([self.hcol_indices[-1]])))
        elif first:
            self.col_indices = np.concatenate((np.array([self.hcol_indices[0]]),
                                               self.col_indices))
        elif last:
            self.col_indices = np.concatenate((self.col_indices,
                                               np.array([self.hcol_indices[-1]])))


    def _extrapolate_rows(self, data):
        """Extrapolate the rows of data, to get the first and last together
        with the data.

        >>> import numpy as np
        >>> lons = np.arange(10).reshape((2, 5), order="F")
        >>> lats = np.arange(10).reshape((2, 5), order="C")
        >>> lines = np.array([2, 7])
        >>> cols = np.array([2, 7, 12, 17, 22])
        >>> hlines = np.arange(10)
        >>> hcols = np.arange(24)
        >>> satint = SatelliteInterpolator((lons, lats), (lines, cols), (hlines, hcols))
        >>> satint._extrapolate_rows(satint.x__)
        array([[ 6381081.08333225,  6381639.66045187,  6372470.10269454,
                 6353590.21586788,  6325042.05851245],
               [ 6370997.        ,  6366146.21816553,  6351605.98629588,
                 6327412.61244969,  6293626.50067273],
               [ 6345786.79166939,  6327412.61244969,  6299445.69529922,
                 6261968.60390423,  6215087.60607344],
               [ 6335702.70833714,  6311919.17016336,  6278581.57890056,
                 6235791.00048604,  6183672.04823372]])
        """

        pos = self.row_indices[:2]
        first_row = _linear_extrapolate(pos,
                                        (data[0, :], data[1, :]),
                                        self.hrow_indices[0])
        pos = self.row_indices[-2:]
        last_row = _linear_extrapolate(pos,
                                       (data[-2, :], data[-1, :]),
                                       self.hrow_indices[-1])

        return np.vstack((np.expand_dims(first_row, 0),
                          data,
                          np.expand_dims(last_row, 0))) 

    def _fill_row_borders(self):
        """Add the first and last rows to the data by extrapolation.

        >>> import numpy as np
        >>> lons = np.arange(20).reshape((4, 5), order="F")
        >>> lats = np.arange(20).reshape((4, 5), order="C")
        >>> lines = np.array([2, 7, 12, 17])
        >>> cols = np.array([2, 7, 12, 17, 22])
        >>> hlines = np.arange(20)
        >>> hcols = np.arange(24)
        >>> satint = SatelliteInterpolator((lons, lats), (lines, cols), (hlines, hcols))
        >>> satint._fill_row_borders()
        >>> satint.x__
        array([[ 6381081.08333225,  6371519.34066148,  6328950.00792935,
                 6253610.69157758,  6145946.19489936],
               [ 6370997.        ,  6354509.6014956 ,  6305151.62592155,
                 6223234.99818839,  6109277.14889072],
               [ 6345786.79166939,  6311985.2535809 ,  6245655.67090206,
                 6147295.76471541,  6017604.5338691 ],
               [ 6270385.16249031,  6219684.08214232,  6137100.75832981,
                 6023313.2794414 ,  5879194.72399075],
               [ 6145476.82098957,  6078546.55734877,  5980676.23854351,
                 5852716.72120069,  5695705.57359808],
               [ 6095513.48438928,  6022091.54743135,  5918106.430629  ,
                 5784478.09790441,  5622309.91344102]])
        >>> satint.row_indices
        array([ 0,  2,  7, 12, 17, 19])
        >>> satint = SatelliteInterpolator((lons, lats), (lines, cols), (hlines, hcols), chunk_size=10)
        >>> satint._fill_row_borders()
        >>> satint.x__
        array([[ 6381081.08333225,  6371519.34066148,  6328950.00792935,
                 6253610.69157758,  6145946.19489936],
               [ 6370997.        ,  6354509.6014956 ,  6305151.62592155,
                 6223234.99818839,  6109277.14889072],
               [ 6345786.79166939,  6311985.2535809 ,  6245655.67090206,
                 6147295.76471541,  6017604.5338691 ],
               [ 6335702.70833714,  6294975.51441502,  6221857.28889426,
                 6116920.07132621,  5980935.48786045],
               [ 6320348.4990906 ,  6276139.09205974,  6199670.56624433,
                 6091551.90273768,  5952590.38414781],
               [ 6270385.16249031,  6219684.08214232,  6137100.75832981,
                 6023313.2794414 ,  5879194.72399075],
               [ 6145476.82098957,  6078546.55734877,  5980676.23854351,
                 5852716.72120069,  5695705.57359808],
               [ 6095513.48438928,  6022091.54743135,  5918106.430629  ,
                 5784478.09790441,  5622309.91344102]])
         >>> satint.row_indices
         array([ 0,  2,  7,  9, 10, 12, 17, 19])
        """
        lines = len(self.hrow_indices)
        chunk_size = self.chunk_size or lines

        x__, y__, z__ = [], [], []
        row_indices = []
        for index in range(0, lines, chunk_size):
            ties = np.argwhere(np.logical_and(self.row_indices >= index,
                                              self.row_indices < index
                                              + chunk_size)).squeeze()
            tiepos = self.row_indices[np.logical_and(self.row_indices >= index,
                                                     self.row_indices < index
                                                     + chunk_size)].squeeze()
            x__.append(self._extrapolate_rows(self.x__[ties, :]))
            y__.append(self._extrapolate_rows(self.y__[ties, :]))
            z__.append(self._extrapolate_rows(self.z__[ties, :]))
            row_indices.append(np.array([self.hrow_indices[index]]))
            row_indices.append(tiepos)
            row_indices.append(np.array([self.hrow_indices[index
                                                           + chunk_size - 1]]))
        self.x__ = np.vstack(x__)
        self.y__ = np.vstack(y__)
        self.z__ = np.vstack(z__)

        self.row_indices = np.concatenate(row_indices)
    
    def _interp(self):
        """Interpolate the cartesian coordinates.
        """
        if np.all(self.hrow_indices == self.row_indices):
            return self._interp1d()
        
        xpoints, ypoints = np.meshgrid(self.hrow_indices,
                                       self.hcol_indices)
        spl = RectBivariateSpline(self.row_indices,
                                  self.col_indices,
                                  self.x__,
                                  s=0,
                                  kx=self.kx_,
                                  ky=self.ky_)

        self.newx = spl.ev(xpoints.ravel(), ypoints.ravel())
        self.newx = self.newx.reshape(xpoints.shape).T

        spl = RectBivariateSpline(self.row_indices,
                                  self.col_indices,
                                  self.y__,
                                  s=0,
                                  kx=self.kx_,
                                  ky=self.ky_)

        self.newy = spl.ev(xpoints.ravel(), ypoints.ravel())
        self.newy = self.newy.reshape(xpoints.shape).T

        spl = RectBivariateSpline(self.row_indices,
                                  self.col_indices,
                                  self.z__,
                                  s=0,
                                  kx=self.kx_,
                                  ky=self.ky_)

        self.newz = spl.ev(xpoints.ravel(), ypoints.ravel())
        self.newz = self.newz.reshape(xpoints.shape).T

    def _interp1d(self):
        """Interpolate in one dimension.
        """
        lines = len(self.hrow_indices)

        self.newx = np.empty((len(self.hrow_indices),
                              len(self.hcol_indices)),
                             self.x__.dtype)

        self.newy = np.empty((len(self.hrow_indices),
                              len(self.hcol_indices)),
                             self.y__.dtype)

        self.newz = np.empty((len(self.hrow_indices),
                              len(self.hcol_indices)),
                             self.z__.dtype)


        for cnt in range(lines):
            tck = splrep(self.col_indices, self.x__[cnt, :], k=self.ky_, s=0)
            self.newx[cnt, :] = splev(self.hcol_indices, tck, der=0)

            tck = splrep(self.col_indices, self.y__[cnt, :], k=self.ky_, s=0)
            self.newy[cnt, :] = splev(self.hcol_indices, tck, der=0)

            tck = splrep(self.col_indices, self.z__[cnt, :], k=self.ky_, s=0)
            self.newz[cnt, :] = splev(self.hcol_indices, tck, der=0)


    def interpolate(self):
        """Do the interpolation, and return resulting longitudes and latitudes.
        """
        self._interp()

        self.longitude = get_lons_from_cartesian(self.newx, self.newy)

        self.latitude = get_lats_from_cartesian(self.newx, self.newy, self.newz)

        return self.longitude, self.latitude

def get_lons_from_cartesian(x__, y__):
    """Get longitudes from cartesian coordinates.
    """
    return rad2deg(arccos(x__ / sqrt(x__ ** 2 + y__ ** 2))) * sign(y__)
    
def get_lats_from_cartesian(x__, y__, z__, thr=0.8):
    """Get latitudes from cartesian coordinates.
    """
    # if we are at low latitudes - small z, then get the
    # latitudes only from z. If we are at high latitudes (close to the poles)
    # then derive the latitude using x and y:

    lats = np.where(np.logical_and(np.less(z__, thr * EARTH_RADIUS), 
                                   np.greater(z__, -1. * thr * EARTH_RADIUS)),
                    90 - rad2deg(arccos(z__/EARTH_RADIUS)),
                    sign(z__) *
                    (90 - rad2deg(arcsin(sqrt(x__ ** 2 + y__ ** 2)
                                         / EARTH_RADIUS))))
    return lats


import unittest

class TestMODIS(unittest.TestCase):
    """Class for system testing the MODIS interpolation.
    """

    def test_5_to_1(self):
        """test the 5km to 1km interpolation facility
        """
        gfilename = \
              "/san1/test/data/modis/MOD03_A12097_174256_2012097175435.hdf"
        filename = \
              "/san1/test/data/modis/MOD021km_A12097_174256_2012097175435.hdf"
        from pyhdf.SD import SD
        from pyhdf.error import HDF4Error
        
        try:
            gdata = SD(gfilename)
            data = SD(filename)
        except HDF4Error:
            print "Failed reading both eos-hdf files %s and %s" % (gfilename, filename)
            return
        
        glats = gdata.select("Latitude")[:]
        glons = gdata.select("Longitude")[:]
    
        lats = data.select("Latitude")[:]
        lons = data.select("Longitude")[:]
        
        tlons, tlats = modis5kmto1km(lons, lats)

        self.assert_(np.allclose(tlons, glons, atol=0.05))
        self.assert_(np.allclose(tlats, glats, atol=0.05))


    def test_1000m_to_250m(self):
        """test the 1 km to 250 meter interpolation facility
        """
        #gfilename = \
        #      "/san1/test/data/modis/MOD03_A12278_113638_2012278145123.hdf"
        gfilename = \
              "/local_disk/src/python-geotiepoints/tests/MOD03_A12278_113638_2012278145123.hdf"
        #result_filename = \
        #      "/san1/test/data/modis/250m_lonlat_results.npz"
        result_filename = \
              "/local_disk/src/python-geotiepoints/tests/250m_lonlat_results.npz"

        from pyhdf.SD import SD
        from pyhdf.error import HDF4Error
        
        try:
            gdata = SD(gfilename)
        except HDF4Error:
            print "Failed reading eos-hdf file %s" % gfilename
            return
        
        lats = gdata.select("Latitude")[0:50, :]
        lons = gdata.select("Longitude")[0:50, :]
    
        verif = np.load(result_filename)
        vlons = verif['lons']
        vlats = verif['lats']
        tlons, tlats = modis1kmto250m(lons, lats)

        self.assert_(np.allclose(tlons, vlons, atol=0.05))
        self.assert_(np.allclose(tlats, vlats, atol=0.05))


if __name__ == "__main__":
    unittest.main()

    



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright (c) 2014 Martin Raspaud

# Author(s):

#   Martin Raspaud <martin.raspaud@smhi.se>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""slerp implementation in numpy
"""


import numpy as np


def tocart(lon, lat):
    rlat = np.deg2rad(lat)
    rlon = np.deg2rad(lon)
    x__ = np.cos(rlat) * np.cos(rlon)
    y__ = np.cos(rlat) * np.sin(rlon)
    z__ = np.sin(rlat)
    return np.dstack((x__, y__, z__))


def toll(arr):
    lat = np.arcsin(arr.take(2, -1))
    lon = np.arctan2(arr.take(1, -1), arr.take(0, -1))
    return np.rad2deg(np.dstack((lon.squeeze(), lat.squeeze())))


def dot(a, b):
    return np.sum(a * b, -1)


def slerp(lon0, lat0, lon1, lat1, t):
    cp0 = tocart(lon0, lat0)
    cp1 = tocart(lon1, lat1)
    omega = np.arccos(dot(cp0, cp1))[:, :, np.newaxis]
    return toll(np.sin((1 - t) * omega) / np.sin(omega) * cp0 +
                np.sin(t * omega) / np.sin(omega) * cp1)



#!/usr/bin/env python

# -*- coding: utf-8 -*-

# Copyright (c) 2014, 2015 Abhay Devasthale

# Author(s):

#   Abhay Devasthale <abhay.devasthale@smhi.se>
#   Adam Dybbroe <adam.dybbroe@smhi.se>
#   Martin Raspaud <martin.raspaud@smhi.se>

# This work was done in the framework of ESA-CCI-Clouds phase I


# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


"""Read a gac file.

Format specification can be found here:
http://www.ncdc.noaa.gov/oa/pod-guide/ncdc/docs/podug/html/c2/sec2-0.htm
http://www.ncdc.noaa.gov/oa/pod-guide/ncdc/docs/podug/html/c3/sec3-1.htm

"""

import datetime
import logging

import numpy as np

import pygac.geotiepoints as gtp
from pygac import gac_io
from pygac.gac_reader import GACReader

LOG = logging.getLogger(__name__)

# common header
header0 = np.dtype([("noaa_spacecraft_identification_code", ">u1"),
                    ("data_type_code", ">u1"),
                    ("start_time", ">u2", (3, )),
                    ("number_of_scans", ">u2"),
                    ("end_time", ">u2", (3, ))])


# until 8 september 1992
header1 = np.dtype([("noaa_spacecraft_identification_code", ">u1"),
                    ("data_type_code", ">u1"),
                    ("start_time", ">u2", (3, )),
                    ("number_of_scans", ">u2"),
                    ("end_time", ">u2", (3, )),
                    ("processing_block_id", "S7"),
                    ("ramp_auto_calibration", ">u1"),
                    ("number_of_data_gaps", ">u2"),
                    ("dacs_quality", ">u1", (6, )),
                    ("calibration_parameter_id", ">i2"),
                    ("dacs_status", ">u1"),
                    ("spare1", ">i1", (5, )),
                    ("data_set_name", "S44"),
                    ("spare2", ">u2", (1568, ))])

header2 = np.dtype([("noaa_spacecraft_identification_code", ">u1"),
                    ("data_type_code", ">u1"),
                    ("start_time", ">u2", (3, )),
                    ("number_of_scans", ">u2"),
                    ("end_time", ">u2", (3, )),
                    ("processing_block_id", "S7"),
                    ("ramp_auto_calibration", ">u1"),
                    ("number_of_data_gaps", ">u2"),
                    ("dacs_quality", ">u1", (6, )),
                    ("calibration_parameter_id", ">i2"),
                    ("dacs_status", ">u1"),
                    ("spare1", ">i1", (5, )),
                    ("data_set_name", "S42"),
                    ("blankfill", "S2"),
                    ("year_of_epoch", ">u2"),
                    ("julian_day_of_epoch", ">u2"),
                    ("millisecond_utc_epoch_time_of_day", ">u4"),
                    # Keplerian orbital elements
                    ("semi_major_axis", ">f8"),
                    ("eccentricity", ">f8"),
                    ("inclination", ">f8"),
                    ("argument_of_perigee", ">f8"),
                    ("right_ascension", ">f8"),
                    ("mean_anomaly", ">f8"),
                    # cartesian inertial true date of elements
                    ("x_component_of_position_vector", ">f8"),
                    ("y_component_of_position_vector", ">f8"),
                    ("z_component_of_position_vector", ">f8"),
                    ("x_dot_component_of_position_vector", ">f8"),
                    ("y_dot_component_of_position_vector", ">f8"),
                    ("z_dot_component_of_position_vector", ">f8"),
                    ("spare2", ">u2", (1516, ))])

header3 = np.dtype([("noaa_spacecraft_identification_code", ">u1"),
                    ("data_type_code", ">u1"),
                    ("start_time", ">u2", (3, )),
                    ("number_of_scans", ">u2"),
                    ("end_time", ">u2", (3, )),
                    ("processing_block_id", "S7"),
                    ("ramp_auto_calibration", ">u1"),
                    ("number_of_data_gaps", ">u2"),
                    ("dacs_quality", ">u1", (6, )),
                    ("calibration_parameter_id", ">i2"),
                    ("dacs_status", ">u1"),
                    ("reserved_for_mounting_and_fixed_attitude_correction_indicator",
                     ">i1"),
                    ("nadir_earth_location_tolerance", ">i1"),
                    ("spare1", ">i1"),
                    ("start_of_data_set_year", ">u2"),
                    ("data_set_name", "S44"),
                    ("year_of_epoch", ">u2"),
                    ("julian_day_of_epoch", ">u2"),
                    ("millisecond_utc_epoch_time_of_day", ">u4"),
                    # Keplerian orbital elements
                    ("semi_major_axis", ">i4"),
                    ("eccentricity", ">i4"),
                    ("inclination", ">i4"),
                    ("argument_of_perigee", ">i4"),
                    ("right_ascension", ">i4"),
                    ("mean_anomaly", ">i4"),
                    # cartesian inertial true date of elements
                    ("x_component_of_position_vector", ">i4"),
                    ("y_component_of_position_vector", ">i4"),
                    ("z_component_of_position_vector", ">i4"),
                    ("x_dot_component_of_position_vector", ">i4"),
                    ("y_dot_component_of_position_vector", ">i4"),
                    ("z_dot_component_of_position_vector", ">i4"),
                    # future use
                    ("yaw_fixed_error_correction", ">i2"),
                    ("roll_fixed_error_correction", ">i2"),
                    ("pitch_fixed_error_correction", ">i2"),
                    ("spare2", ">u2", (1537, ))])


scanline = np.dtype([("scan_line_number", ">i2"),
                     ("time_code", ">u2", (3, )),
                     #("time_code", ">u1", (6, )),
                     ("quality_indicators", ">u4"),
                     ("calibration_coefficients", ">i4", (10, )),
                     ("number_of_meaningful_zenith_angles_and_earth_location_appended",
                      ">u1"),
                     ("solar_zenith_angles", "i1", (51, )),
                     ("earth_location", ">i2", (102, )),
                     ("telemetry", ">u4", (35, )),
                     ("sensor_data", ">u4", (682, )),
                     ("add_on_zenith", ">u2", (10, )),
                     ("clock_drift_delta", ">u2"),
                     ("spare3", "u2", (11, ))])


class PODReader(GACReader):

    spacecrafts_orbital = {2: 'noaa 6',
                           4: 'noaa 7',
                           6: 'noaa 8',
                           7: 'noaa 9',
                           8: 'noaa 10',
                           1: 'noaa 11',
                           5: 'noaa 12',
                           3: 'noaa 14',
                           }
    spacecraft_names = {2: 'noaa6',
                        4: 'noaa7',
                        6: 'noaa8',
                        7: 'noaa9',
                        8: 'noaa10',
                        1: 'noaa11',
                        5: 'noaa12',
                        3: 'noaa14',
                        }

    def read(self, filename):
        # choose the right header depending on the date
        with open(filename) as fd_:
            head = np.fromfile(fd_, dtype=header0, count=1)[0]
            year = head["start_time"][0] >> 9
            year = np.where(year > 75, year + 1900, year + 2000)
            jday = (head["start_time"][0] & 0x1FF)

            start_date = (datetime.date(year, 1, 1) +
                          datetime.timedelta(days=int(jday) - 1))

            if start_date < datetime.date(1992, 9, 8):
                header = header1
            elif start_date <= datetime.date(1994, 11, 15):
                header = header2
            else:
                header = header3

        with open(filename) as fd_:
            self.head = np.fromfile(fd_, dtype=header, count=1)[0]
            scans = np.fromfile(fd_,
                                dtype=scanline,
                                count=self.head["number_of_scans"])

        # cleaning up the data
        min_scanline_number = np.amin(
            np.absolute(scans["scan_line_number"][:]))
        if scans["scan_line_number"][0] == scans["scan_line_number"][-1] + 1:
            while scans["scan_line_number"][0] != min_scanline_number:
                scans = np.roll(scans, -1)
        else:
            while scans["scan_line_number"][0] != min_scanline_number:
                scans = scans[1:]

        self.scans = scans[scans["scan_line_number"] != 0]

        self.spacecraft_id = self.head["noaa_spacecraft_identification_code"]
        self.spacecraft_name = self.spacecraft_names[self.spacecraft_id]
        LOG.info(
            "Reading %s data", self.spacecrafts_orbital[self.spacecraft_id])

        return self.head, self.scans

    def get_times(self):
        if self.utcs is None:
            year = self.scans["time_code"][:, 0] >> 9
            year = np.where(year > 75, year + 1900, year + 2000)
            jday = (self.scans["time_code"][:, 0] & 0x1FF)
            msec = ((np.uint32(self.scans["time_code"][:, 1] & 2047) << 16) |
                    (np.uint32(self.scans["time_code"][:, 2])))

            # print jday[10150:10171]
            # print msec[10150:10171]
            jday = np.where(np.logical_or(jday < 1, jday > 366),
                            np.median(jday), jday)
            if_wrong_jday = np.ediff1d(jday, to_begin=0)
            jday = np.where(if_wrong_jday < 0, max(jday), jday)

            if_wrong_msec = np.where(msec < 1)
            if_wrong_msec = if_wrong_msec[0]
            if len(if_wrong_msec) > 0:
                if if_wrong_msec[0] != 0:
                    msec = msec[0] + 0.5 * 1000.0 * \
                        (self.scans["scan_line_number"] - 1)
                else:
                    msec = np.median(msec - 0.5 * 1000.0 *
                                     (self.scans["scan_line_number"] - 1))

            if_wrong_msec = np.ediff1d(msec, to_begin=0)
            #msec = np.where(np.logical_or(if_wrong_msec<-1000, if_wrong_msec>1000), msec[0] + 0.5 * 1000.0 * (self.scans["scan_line_number"] - 1), msec)
            msec = np.where(np.logical_and(np.logical_or(if_wrong_msec < -1000, if_wrong_msec > 1000),
                                           if_wrong_jday != 1), msec[0] + 0.5 * 1000.0 * (self.scans["scan_line_number"] - 1), msec)

            self.utcs = (((year - 1970).astype('datetime64[Y]')
                          + (jday - 1).astype('timedelta64[D]')).astype('datetime64[ms]')
                         + msec.astype('timedelta64[ms]'))
            self.times = self.utcs.astype(datetime.datetime)

            # print if_wrong_jday[10150:10171];
            # print if_wrong_msec[10150:10171];
            # print jday[10150:10171]
            # print msec[10150:10171]
            # print self.utcs[10150:10171].astype(datetime.datetime)
            # checking if year value is out of valid range
            if_wrong_year = np.where(np.logical_or(year < 1978, year > 2015))
            if_wrong_year = if_wrong_year[0]
            if len(if_wrong_year) > 0:
                # if the first scanline has valid time stamp
                if if_wrong_year[0] != 0:
                    year = year[0]
                    jday = jday[0]
                    msec = msec[0] + 0.5 * 1000.0 * \
                        (self.scans["scan_line_number"] - 1)
                    self.utcs = (((year - 1970).astype('datetime64[Y]')
                                  + (jday - 1).astype('timedelta64[D]')).astype('datetime64[ms]')
                                 + msec.astype('timedelta64[ms]'))
                    self.times = self.utcs.astype(datetime.datetime)
                # Otherwise use median time stamp
                else:
                    year = np.median(year)
                    jday = np.median(jday)
                    msec = np.median(msec - 0.5 * 1000.0 *
                                     (self.scans["scan_line_number"] - 1))
                    self.utcs = (((year - 1970).astype('datetime64[Y]')
                                  + (jday - 1).astype('timedelta64[D]')).astype('datetime64[ms]')
                                 + msec.astype('timedelta64[ms]'))
                    self.times = self.utcs.astype(datetime.datetime)

        return self.utcs

    def adjust_clock_drift(self):
        """Adjust the geolocation to compensate for the clock error.

        TODO: bad things might happen when scanlines are skipped.
        """
        tic = datetime.datetime.now()
        self.get_times()
        from pygac.clock_offsets_converter import get_offsets
        try:
            offset_times, clock_error = get_offsets(self.spacecraft_name)
        except KeyError:
            LOG.info("No clock drift info available for %s",
                     self.spacecraft_name)
        else:
            offset_times = np.array(offset_times, dtype='datetime64[ms]')
            offsets = np.interp(self.utcs.astype(np.uint64),
                                offset_times.astype(np.uint64),
                                clock_error)
            LOG.info("Adjusting for clock drift of %s to %s",
                     str(min(offsets)),
                     str(max(offsets)))
            self.times = (self.utcs +
                          offsets.astype('timedelta64[s]')).astype(datetime.datetime)
            offsets *= -2

            int_offsets = np.floor(offsets).astype(np.int)

            # filling out missing geolocations with computation from pyorbital.
            line_indices = (self.scans["scan_line_number"]
                            + int_offsets)

            missed = sorted((set(line_indices) |
                             set(line_indices + 1))
                            - set(self.scans["scan_line_number"]))

            min_idx = min(line_indices)
            max_idx = max(max(line_indices),
                          max(self.scans["scan_line_number"] - min_idx)) + 1
            idx_len = max_idx - min_idx + 2

            complete_lons = np.zeros((idx_len, 409), dtype=np.float) * np.nan
            complete_lats = np.zeros((idx_len, 409), dtype=np.float) * np.nan

            complete_lons[self.scans["scan_line_number"] - min_idx] = self.lons
            complete_lats[self.scans["scan_line_number"] - min_idx] = self.lats

            missed_utcs = ((np.array(missed) - 1) * np.timedelta64(500, "ms")
                           + self.utcs[0])

            mlons, mlats = self.compute_lonlat(missed_utcs, True)

            complete_lons[missed - min_idx] = mlons
            complete_lats[missed - min_idx] = mlats

            from pygac.slerp import slerp
            off = offsets - np.floor(offsets)
            res = slerp(complete_lons[line_indices - min_idx, :],
                        complete_lats[line_indices - min_idx, :],
                        complete_lons[line_indices - min_idx + 1, :],
                        complete_lats[line_indices - min_idx + 1, :],
                        off[:, np.newaxis, np.newaxis])

            self.lons = res[:, :, 0]
            self.lats = res[:, :, 1]
            self.utcs += offsets.astype('timedelta64[s]')

        toc = datetime.datetime.now()
        LOG.debug("clock drift adjustment took %s", str(toc - tic))

    def get_lonlat(self):
        # interpolating lat-on points using PYTROLL geotiepoints
        arr_lat = self.scans["earth_location"][:, 0::2] / 128.0
        arr_lon = self.scans["earth_location"][:, 1::2] / 128.0

        self.lons, self.lats = gtp.Gac_Lat_Lon_Interpolator(arr_lon, arr_lat)
        return self.lons, self.lats

    def get_telemetry(self):
        number_of_scans = self.scans["telemetry"].shape[0]
        decode_tele = np.zeros((int(number_of_scans), 105))
        decode_tele[:, ::3] = (self.scans["telemetry"] >> 20) & 1023
        decode_tele[:, 1::3] = (self.scans["telemetry"] >> 10) & 1023
        decode_tele[:, 2::3] = self.scans["telemetry"] & 1023

        prt_counts = np.mean(decode_tele[:, 17:20], axis=1)

        # getting ICT counts

        ict_counts = np.zeros((int(number_of_scans), 3))
        ict_counts[:, 0] = np.mean(decode_tele[:, 22:50:3], axis=1)
        ict_counts[:, 1] = np.mean(decode_tele[:, 23:51:3], axis=1)
        ict_counts[:, 2] = np.mean(decode_tele[:, 24:52:3], axis=1)

        # getting space counts

        space_counts = np.zeros((int(number_of_scans), 3))
        space_counts[:, 0] = np.mean(decode_tele[:, 54:100:5], axis=1)
        space_counts[:, 1] = np.mean(decode_tele[:, 55:101:5], axis=1)
        space_counts[:, 2] = np.mean(decode_tele[:, 56:102:5], axis=1)

        return prt_counts, ict_counts, space_counts

    def get_corrupt_mask(self):

        # corrupt scanlines

        mask = ((self.scans["quality_indicators"] >> 31) |
                ((self.scans["quality_indicators"] << 4) >> 31) |
                ((self.scans["quality_indicators"] << 5) >> 31))

        number_of_scans = self.scans["telemetry"].shape[0]
        qual_flags = np.zeros((int(number_of_scans), 7))
        qual_flags[:, 0] = self.scans["scan_line_number"]
        qual_flags[:, 1] = (self.scans["quality_indicators"] >> 31)
        qual_flags[:, 2] = ((self.scans["quality_indicators"] << 4) >> 31)
        qual_flags[:, 3] = ((self.scans["quality_indicators"] << 5) >> 31)
        qual_flags[:, 4] = ((self.scans["quality_indicators"] << 13) >> 31)
        qual_flags[:, 5] = ((self.scans["quality_indicators"] << 14) >> 31)
        qual_flags[:, 6] = ((self.scans["quality_indicators"] << 15) >> 31)

        return mask.astype(bool), qual_flags


def main(filename, start_line, end_line):
    tic = datetime.datetime.now()
    reader = PODReader()
    reader.read(filename)
    reader.get_lonlat()
    reader.adjust_clock_drift()
    channels = reader.get_calibrated_channels()
    sat_azi, sat_zen, sun_azi, sun_zen, rel_azi = reader.get_angles()

    mask, qual_flags = reader.get_corrupt_mask()
    if (np.all(mask)):
        print "ERROR: All data is masked out. Stop processing"
        raise ValueError("All data is masked out.")

    gac_io.save_gac(reader.spacecraft_name,
                    reader.utcs,
                    reader.lats, reader.lons,
                    channels[:, :, 0], channels[:, :, 1],
                    np.ones_like(channels[:, :, 0]) * -1,
                    channels[:, :, 2],
                    channels[:, :, 3],
                    channels[:, :, 4],
                    sun_zen, sat_zen, sun_azi, sat_azi, rel_azi,
                    mask, qual_flags, start_line, end_line)
    LOG.info("pygac took: %s", str(datetime.datetime.now() - tic))


if __name__ == "__main__":
    import sys
    main(sys.argv[1], sys.argv[2], sys.argv[3])



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright (c) 2014, 2015 Martin Raspaud

# Author(s):

#   Martin Raspaud <martin.raspaud@smhi.se>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Generic reader for GAC data. Can't be used as is, has to be subclassed to add
specific read functions.
"""
import numpy as np
from pygac import CONFIG_FILE
import ConfigParser
import os
import logging
from pyorbital.orbital import Orbital
from pyorbital import astronomy
import datetime
from pygac.gac_calibration import calibrate_solar, calibrate_thermal

LOG = logging.getLogger(__name__)


class GACReader(object):

    def __init__(self):
        self.head = None
        self.scans = None
        self.spacecraft_name = None
        self.spacecraft_id = None
        self.utcs = None
        self.lats = None
        self.lons = None
        self.times = None
        self.tle_lines = None

    def get_counts(self):
        packed_data = self.scans["sensor_data"]
        gac_counts = np.zeros((len(self.scans), 409 * 5))
        gac_counts[:, 0::3] = (packed_data >> 20) & 1023
        gac_counts[:, 1::3] = (packed_data >> 10) & 1023
        gac_counts[:, 2::3] = (packed_data & 1023)[:, :-1]
        gac_counts = gac_counts.reshape((-1, 409, 5))
        try:
            switch = self.get_ch3_switch()
        except AttributeError:
            return gac_counts
        else:
            channels = np.zeros((len(self.scans), 409, 6),
                                dtype=gac_counts.dtype)
            channels[:, :, :2] = gac_counts[:, :, :2]
            channels[:, :, -2:] = gac_counts[:, :, -2:]
            channels[:, :, 2][switch == 1] = gac_counts[:, :, 2][switch == 1]
            channels[:, :, 3][switch == 0] = gac_counts[:, :, 2][switch == 0]
            return channels

    def get_times(self):
        raise NotImplementedError

    def compute_lonlat(self, utcs=None, clock_drift_adjust=True):
        tle1, tle2 = self.get_tle_lines()

        scan_points = np.arange(3.5, 2048, 5)

        if utcs is None:
            utcs = self.get_times()

        # adjusting clock for drift
        tic = datetime.datetime.now()
        if clock_drift_adjust:
            from pygac.clock_offsets_converter import get_offsets
            try:
                offset_times, clock_error = get_offsets(self.spacecraft_name)
            except KeyError:
                LOG.info("No clock drift info available for %s",
                         self.spacecraft_name)
            else:
                offset_times = np.array(offset_times, dtype='datetime64[ms]')
                offsets = np.interp(utcs.astype(np.uint64),
                                    offset_times.astype(np.uint64),
                                    clock_error)
                utcs -= (offsets * 1000).astype('timedelta64[ms]')

        t = utcs[0].astype(datetime.datetime)

        if "constant_yaw_attitude_error" in self.head.dtype.fields:
            rpy = np.deg2rad([self.head["constant_roll_attitude_error"] / 1e3,
                              self.head["constant_pitch_attitude_error"] / 1e3,
                              self.head["constant_yaw_attitude_error"] / 1e3])
        else:
            rpy = [0, 0, 0]

        LOG.info("Using rpy: %s", str(rpy))

        from pyorbital.geoloc_instrument_definitions import avhrr_gac
        from pyorbital.geoloc import compute_pixels, get_lonlatalt
        sgeom = avhrr_gac(utcs.astype(datetime.datetime), scan_points, 55.385)
        s_times = sgeom.times(t)

        pixels_pos = compute_pixels((tle1, tle2), sgeom, s_times, rpy)
        pos_time = get_lonlatalt(pixels_pos, s_times)

        toc = datetime.datetime.now()

        LOG.warning("Computation of geolocation: %s", str(toc - tic))

        lons, lats = pos_time[:2]

        return lons.reshape(-1, 409), lats.reshape(-1, 409)

    def get_calibrated_channels(self):
        channels = self.get_counts()
        self.get_times()
        year = self.times[0].year
        delta = self.times[0].date() - datetime.date(year, 1, 1)
        jday = delta.days + 1

        # Earth-Sun distance correction factor
        corr = 1.0 - 0.0334 * np.cos(2.0 * np.pi * (jday - 2) / 365.25)

        # how many reflective channels are there ?
        tot_ref = channels.shape[2] - 3

        channels[:, :, 0:tot_ref] = calibrate_solar(channels[:, :, 0:tot_ref],
                                                    np.arange(tot_ref),
                                                    year, jday,
                                                    self.spacecraft_name,
                                                    corr)
        prt, ict, space = self.get_telemetry()
        for chan in [3, 4, 5]:
            channels[:, :, chan - 6] = calibrate_thermal(
                channels[:, :, chan - 6],
                prt,
                ict[:, chan - 3],
                space[:, chan - 3],
                self.scans["scan_line_number"],
                chan,
                self.spacecraft_name)
        return channels

    def get_tle_file(self):
        conf = ConfigParser.ConfigParser()
        try:
            conf.read(CONFIG_FILE)
        except ConfigParser.NoSectionError:
            LOG.exception('Failed reading configuration file: %s',
                          str(CONFIG_FILE))
            raise

        values = {"satname": self.spacecraft_name, }
        options = {}
        for option, value in conf.items('tle', raw=True):
            options[option] = value
        tle_filename = os.path.join(options['tledir'],
                                    options["tlename"] % values)
        LOG.info('TLE filename = ' + str(tle_filename))
        with open(tle_filename, 'r') as fp_:
            return fp_.readlines()

    def get_tle_lines(self):
        if self.tle_lines is not None:
            return self.tle_lines
        tle_data = self.get_tle_file()
        tm = self.times[0]
        sdate = (int(tm.strftime("%Y%j")) +
                 (tm.hour +
                  (tm.minute +
                   (tm.second +
                    tm.microsecond / 1000000.0) / 60.0) / 60.0) / 24.0)

        dates = np.array([float(line[18:32]) for line in tle_data[::2]])
        dates = np.where(dates > 50000, dates + 1900000, dates + 2000000)

        iindex = np.searchsorted(dates, sdate)

        if ((iindex == 0 and abs(sdate - dates[0]) > 7) or
                (iindex == len(dates) - 1 and abs(sdate - dates[-1]) > 7)):
            raise IndexError(
                "Can't find tle data for %s on the %s" %
                (self.spacecraft_name, tm.isoformat()))

        if abs(sdate - dates[iindex - 1]) < abs(sdate - dates[iindex]):
            iindex -= 1

        if abs(sdate - dates[iindex]) > 3:
            LOG.warning("Found TLE data for %f that is %f days appart",
                        sdate, abs(sdate - dates[iindex]))
        else:
            LOG.debug("Found TLE data for %f that is %f days appart",
                      sdate, abs(sdate - dates[iindex]))

        tle1 = tle_data[iindex * 2]
        tle2 = tle_data[iindex * 2 + 1]
        self.tle_lines = tle1, tle2
        return tle1, tle2

    def get_angles(self):
        tle1, tle2 = self.get_tle_lines()
        orb = Orbital(self.spacecrafts_orbital[self.spacecraft_id],
                      line1=tle1, line2=tle2)

        sat_azi, sat_elev = orb.get_observer_look(self.times[:, np.newaxis],
                                                  self.lons, self.lats, 0)

        sat_zenith = 90 - sat_elev

        sun_zenith = astronomy.sun_zenith_angle(self.times[:, np.newaxis],
                                                self.lons, self.lats)

        alt, sun_azi = astronomy.get_alt_az(self.times[:, np.newaxis],
                                            self.lons, self.lats)
        del alt
        sun_azi = np.rad2deg(sun_azi)
        sun_azi = np.where(sun_azi < 0, sun_azi + 180, sun_azi - 180)

        rel_azi = abs(sat_azi - sun_azi)
        rel_azi = np.where(rel_azi > 180.0, 360.0 - rel_azi, rel_azi)

        return sat_azi, sat_zenith, sun_azi, sun_zenith, rel_azi



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright (c) 2014 Abhay Devasthale, Adam.Dybbroe, Martin Raspaud

# Author(s):

#   Adam.Dybbroe <a000680@c14526.ad.smhi.se>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


try:
    with open("./README", "r") as fd:
        long_description = fd.read()
except IOError:
    long_description = ""


from setuptools import setup
import imp

version = imp.load_source('pygac.version', 'pygac/version.py')

setup(name='pygac',
      version="v0.1.0",
      description='NOAA AVHRR GAC reader and calibration',
      author='Abhay Devasthale',
      author_email='adam.dybbroe@smhi.se',
      classifiers=["Development Status :: 4 - Beta",
                   "Intended Audience :: Science/Research",
                   "License :: OSI Approved :: GNU General Public License v3 " +
                   "or later (GPLv3+)",
                   "Operating System :: OS Independent",
                   "Programming Language :: Python",
                   "Topic :: Scientific/Engineering"],
      url="https://github.com/adybbroe/pygac",
      long_description=long_description,
      license='GPLv3',

      packages=['pygac'],

      # Project should use reStructuredText, so ensure that the docutils get
      # installed or upgraded on the target machine
      install_requires=['docutils>=0.3',
                        'numpy>=1.8.0', 'pyorbital>=v0.3.2',
                        'h5py'],
      extras_require={'geolocation interpolation': ['python-geotiepoints'],
                      },
      scripts=[],
      data_files=[('etc', ['etc/pygac.cfg.template']),
                  ('gapfilled_tles', ['gapfilled_tles/TLE_noaa16.txt'])],
      test_suite="pygac.tests.suite",
      tests_require=[],

      zip_safe=False
      )



#!/usr/bin/env python

# -*- coding: utf-8 -*-

# Copyright (c) 2014 Abhay Devasthale, Martin Raspaud

# Author(s):

#   Abhay Devasthale <abhay.devasthale@smhi.se>
#   Martin Raspaud   <martin.raspaud@smhi.se>

# This work was done in the framework of ESA-CCI-Clouds phase I


# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY
# without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.	See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


import numpy as np

MISSING_DATA = -32001


# calibrates solar channels of AVHRR


def calibrate_solar_pod(counts, year, jday, spacecraft_id, channel3_switch, corr, number_of_data_records):

    # setting up calibration coefficients

    if spacecraft_id == 4:
        #/*noaa-7*/
        Cdark1 = 36.0
        Cdark2 = 37.0
        Cdark3 = 39.0

        al1 = 0.111206
        bl1 = 8.85810
        cl1 = -1.11938

        ah1 = 0.111206
        bh1 = 8.85810
        ch1 = -1.11938

        al2 = 0.107664
        bl2 = 20.7843
        cl2 = -4.32167

        ah2 = 0.107664
        bh2 = 20.7843
        ch2 = -4.32167

        al3 = 0.0000
        bl3 = 0.00
        cl3 = 0.00

        ah3 = 0.0000
        bh3 = 0.00
        ch3 = 0.00

        Ldate = 1981.4764

    elif spacecraft_id == 7:
        #/*noaa-9*/
        Cdark1 = 38.0
        Cdark2 = 40.0
        Cdark3 = 38.0

        al1 = 0.109312
        bl1 = 5.66526
        cl1 = 0.271520

        ah1 = 0.109312
        bh1 = 5.66526
        ch1 = 0.271520

        al2 = 0.116010
        bl2 = 3.93889
        cl2 = -0.258047

        ah2 = 0.116010
        bh2 = 3.93889
        ch2 = -0.258047

        al3 = 0.0000
        bl3 = 0.00
        cl3 = 0.00

        ah3 = 0.0000
        bh3 = 0.00
        ch3 = 0.00

        Ldate = 1984.9480

    elif spacecraft_id == 8:
        #/*noaa-10*/
        Cdark1 = 39.44
        Cdark2 = 39.40
        Cdark3 = 37.51

        al1 = 0.114803
        bl1 = 3.43456
        cl1 = -0.384635

        ah1 = 0.114803
        bh1 = 3.43456
        ch1 = -0.384635

        al2 = 0.132682
        bl2 = 0.00926103
        cl2 = 0.137333

        ah2 = 0.132682
        bh2 = 0.00926103
        ch2 = 0.137333

        al3 = 0.0
        bl3 = 0.00
        cl3 = 0.0

        ah3 = 0.0
        bh3 = 0.00
        ch3 = 0.0

        Ldate = 1986.712

    elif spacecraft_id == 1:
        #/*noaa-11*/
        Cdark1 = 40.0
        Cdark2 = 40.0
        Cdark3 = 40.0

        al1 = 0.114786
        bl1 = -1.82974
        cl1 = 0.412551

        ah1 = 0.114786
        bh1 = -1.82974
        ch1 = 0.412551

        al2 = 0.115219
        bl2 = -1.39495
        cl2 = 0.236978

        ah2 = 0.115219
        bh2 = -1.39495
        ch2 = 0.236978

        al3 = 0.0000
        bl3 = 0.00
        cl3 = 0.00

        ah3 = 0.0000
        bh3 = 0.00
        ch3 = 0.00

        Ldate = 1988.7064

    elif spacecraft_id == 5:
        # /*noaa-12*/
        Cdark1 = 41.0
        Cdark2 = 40.0
        Cdark3 = 40.0

        al1 = 0.118059
        bl1 = 4.88848
        cl1 = -0.347769

        ah1 = 0.118059
        bh1 = 4.88848
        ch1 = -0.347769

        al2 = 0.135684
        bl2 = 4.67098
        cl2 = -0.399447

        ah2 = 0.135684
        bh2 = 4.67098
        ch2 = -0.399447

        al3 = 0.000
        bl3 = 0.00
        cl3 = 0.00

        ah3 = 0.000
        bh3 = 0.00
        ch3 = 0.00

        Ldate = 1991.3669

    elif spacecraft_id == 3:
        # /*noaa-14*/
        Cdark1 = 41.0
        Cdark2 = 41.0
        Cdark3 = 39.0

        al1 = 0.121408
        bl1 = 4.07313
        cl1 = -0.375215

        ah1 = 0.121408
        bh1 = 4.07313
        ch1 = -0.375215

        al2 = 0.144515
        bl2 = 0.885331
        cl2 = 0.233727

        ah2 = 0.144515
        bh2 = 0.885331
        ch2 = 0.233727

        al3 = 0.0000
        bl3 = 0.00
        cl3 = 0.00

        ah3 = 0.0000
        bh3 = 0.00
        ch3 = 0.00

        Ldate = 1994.9699

    else:
        print "wrong satellite id - exit"

        sys.exit(0)

    print 'year, jday, spacecraft-id, launch date - ', year, jday, spacecraft_id, Ldate

    t = (year + jday / 365.0) - Ldate
    # channel 1
    stl = 0.0
    sth = 0.0
    cindex = 0
    raw_counts = 0

    #r1=np.zeros((number_of_data_records, 409))

    raw_counts = counts[:, 0::5]

    stl = (al1 * (100.0 + bl1 * t + cl1 * t * t)) / 100.0
    # sth=(ah1*(100.0+bh1*t+ch1*t*t))/100.0;
    r1 = (raw_counts - Cdark1) * stl
    # cindex=np.where(raw_counts>Cs1);
    # r1[cindex]=(Cs1-Cdark1)*stl+(raw_counts[cindex]-Cs1)*sth;
    r1 = r1 * corr

    # channel 2
    stl = 0.0
    sth = 0.0
    cindex = 0
    raw_counts = 0

    #r2=np.zeros((number_of_data_records, 409));
    raw_counts = counts[:, 1::5]
    stl = (al2 * (100.0 + bl2 * t + cl2 * t * t)) / 100.0
    # sth=(ah2*(100.0+bh2*t+ch2*t*t))/100.0;
    r2 = (raw_counts - Cdark2) * stl
    # cindex=np.where(raw_counts>Cs2);
    # r2[cindex]=(Cs2-Cdark2)*stl+(raw_counts[cindex]-Cs2)*sth;
    r2 = r2 * corr

    # channel 3a (named as channel 6 in the output HDF5 file)
    r3 = np.zeros((number_of_data_records, raw_counts.shape[1]))
    r3[:, :] = MISSING_DATA

    return r1, r2, r3

 # calibrates thermal channels of AVHRR


def calibrate_thermal_pod(raw_counts, prt, ict, space, number_of_data_records, spacecraft_id, channel, line_numbers):

    if spacecraft_id == 4:
        # /*noaa-7*/
        d10 = 277.099
        d11 = 5.048E-2
        d12 = 2.823E-6
        d20 = 276.734
        d21 = 5.069E-2
        d22 = 2.493E-6
        d30 = 276.876
        d31 = 5.148E-2
        d32 = 1.040E-6
        d40 = 276.160
        d41 = 5.128E-2
        d42 = 1.414E-6
        if(channel == 3):
            Ns = 0.0
            cWavenumber = 2684.5233
            A = 1.94882690
            B = 1.0 / 1.0029260
            b1 = 1.0
            b2 = 0.0
            b0 = 0.0
        elif(channel == 4):
            Ns = -5.16
            cWavenumber = 928.23757
            A = 0.52807997
            B = 1.0 / 1.0014039
            b1 = 0.89783
            b2 = 0.0004819
            b0 = 5.25
        elif(channel == 5):
            Ns = -4.28
            cWavenumber = 841.52137
            A = 0.40557027
            B = 1.0 / 1.0011789
            b1 = 0.93683
            b2 = 0.0002425
            b0 = 3.93

    elif spacecraft_id == 7:
             # /*noaa-9*/
        d10 = 277.018000
        d11 = 0.051280
        d12 = 0.000000E-00
        d20 = 276.750000
        d21 = 0.051280
        d22 = 0.000000E-00
        d30 = 276.862000
        d31 = 0.051280
        d32 = 0.000000E-00
        d40 = 276.546000
        d41 = 0.051280
        d42 = 0.000000E-00
        if(channel == 3):
            Ns = 0.0
            cWavenumber = 2690.0451
            A = 1.8832662
            B = 1.0 / 1.0028978
            b1 = 1.0
            b2 = 0.0
            b0 = 0.0
        elif(channel == 4):
            Ns = -5.530
            cWavenumber = 930.50230
            A = 0.5115335
            B = 1.0 / 1.0013570
            b1 = 0.88643
            b2 = 0.0006033
            b0 = 5.24
        elif(channel == 5):
            Ns = -3.06
            cWavenumber = 845.75000
            A = 0.3882150
            B = 1.0 / 1.0011210
            b1 = 0.95311
            b2 = 0.0002198
            b0 = 2.42

    elif spacecraft_id == 8:
             # /*noaa-10*/
        d10 = 276.659
        d11 = 0.051275
        d12 = 1.363e-06
        d20 = 276.659
        d21 = 0.051275
        d22 = 1.363e-06
        d30 = 276.659
        d31 = 0.051275
        d32 = 1.363e-06
        d40 = 276.659
        d41 = 0.051275
        d42 = 1.363e-06
        if(channel == 3):
            Ns = 0.0
            cWavenumber = 2672.6392
            A = 1.80168640
            B = 1.0 / 1.0026383
            b1 = 1.0
            b2 = 0.0
            b0 = 0.0
        elif(channel == 4):
            Ns = 0.0
            cWavenumber = 910.51930
            A = 0.45707890
            B = 1.0 / 1.0012338
            b1 = 1.0
            b2 = 0.0
            b0 = 0.0
        elif(channel == 5):
            Ns = 0.0
            cWavenumber = 910.51930
            A = 0.45707890
            B = 1.0 / 1.0012338
            b1 = 1.0
            b2 = 0.0
            b0 = 0.0

    elif spacecraft_id == 1:
             # /*noaa-11*/
        d10 = 276.597
        d11 = 0.051275
        d12 = 1.363e-06
        d20 = 276.597
        d21 = 0.051275
        d22 = 1.363e-06
        d30 = 276.597
        d31 = 0.051275
        d32 = 1.363e-06
        d40 = 276.597
        d41 = 0.051275
        d42 = 1.363e-06
        if(channel == 3):
            Ns = 0.0
            cWavenumber = 2680.05
            A = 1.738973
            B = 1.0 / 1.003354
            b1 = 1.0
            b2 = 0.0
            b0 = 0.0
        elif(channel == 4):
            Ns = -8.055
            cWavenumber = 927.462
            A = 0.321199
            B = 1.0 / 1.001213
            b1 = 0.84120
            b2 = 0.0008739
            b0 = 7.21
        elif(channel == 5):
            Ns = -3.51
            cWavenumber = 840.746
            A = 0.048652
            B = 1.0 / 1.000664
            b1 = 0.94598
            b2 = 0.0002504
            b0 = 2.92

    elif spacecraft_id == 5:
             # /*noaa-12*/
        d10 = 276.597
        d11 = 0.051275
        d12 = 1.363e-06
        d20 = 276.597
        d21 = 0.051275
        d22 = 1.363e-06
        d30 = 276.597
        d31 = 0.051275
        d32 = 1.363e-06
        d40 = 276.597
        d41 = 0.051275
        d42 = 1.363e-06
        if(channel == 3):
            Ns = 0.0
            cWavenumber = 2651.7708
            A = 1.90527390
            B = 1.0 / 1.0030100
            b1 = 1.0
            b2 = 0.0
            b0 = 0.0
        elif(channel == 4):
            Ns = -5.510
            cWavenumber = 922.36261
            A = 0.63404209
            B = 1.0 / 1.0017076
            b1 = 0.88929
            b2 = 0.0005968
            b0 = 5.11
        elif(channel == 5):
            Ns = -2.51
            cWavenumber = 838.02678
            A = 0.41086587
            B = 1.0 / 1.0012010
            b1 = 0.96299
            b2 = 0.0001775
            b0 = 1.91

    elif spacecraft_id == 3:
             # /*noaa-14*/
        d10 = 276.597
        d11 = 0.051275
        d12 = 1.363e-06
        d20 = 276.597
        d21 = 0.051275
        d22 = 1.363e-06
        d30 = 276.597
        d31 = 0.051275
        d32 = 1.363e-06
        d40 = 276.597
        d41 = 0.051275
        d42 = 1.363e-06
        if(channel == 3):
            Ns = 0.0069
            cWavenumber = 2654.25
            A = 1.885330
            B = 1.0 / 1.003839
            b1 = 1.00359
            b2 = 0.0
            b0 = -0.0031
        elif(channel == 4):
            Ns = -4.05
            cWavenumber = 928.349
            A = 0.308384
            B = 1.0 / 1.001443
            b1 = 0.92378
            b2 = 0.000382
            b0 = 3.72
        elif(channel == 5):
            Ns = -2.29
            cWavenumber = 833.040
            A = 0.022171
            B = 1.0 / 1.000538
            b1 = 0.96194
            b2 = 0.0001742
            b0 = 2.00
    else:
        print "wrong satellite id - exit"
        sys.exit(0)

    columns = raw_counts.shape[1]

    # adjustment and preparation for calculating four PRT temperatures

    izeros = np.where(prt <= 50)
    izeros = izeros[0]

    inonzeros = np.where(prt > 50)
    inonzeros = inonzeros[0]

    offset = 0

    for i, prt_val in enumerate(prt):
        if prt_val < 0:
            offset = i
            break

    iprt = (line_numbers - 1 - offset) % 5

    tprt = np.zeros((int(number_of_data_records)))
    iones = np.where(iprt == 1)
    itwos = np.where(iprt == 2)
    ithrees = np.where(iprt == 3)
    ifours = np.where(iprt == 4)
    prt = prt.astype(float)
    tprt[iones] = d10 + d11 * prt[iones] + d12 * (prt[iones] * prt[iones])
    tprt[itwos] = d20 + d21 * prt[itwos] + d22 * (prt[itwos] * prt[itwos])
    tprt[ithrees] = d30 + d31 * prt[ithrees] + \
        d32 * (prt[ithrees] * prt[ithrees])
    tprt[ifours] = d40 + d41 * prt[ifours] + d42 * (prt[ifours] * prt[ifours])

    izeros = np.where(iprt == 0)
    izeros = izeros[0]

    inonzeros = np.where(iprt > 0)
    inonzeros = inonzeros[0]
    temp_prt = np.interp(izeros, inonzeros, tprt[inonzeros])
    tprt[izeros] = temp_prt

    # convolving and smoothing PRT, ICT and SPACE values

    if number_of_data_records > 51:
        window = 51
        # note that the window size has to be an odd number and greater than 2
    else:
        window = 3

    weighting_function = np.ones(window, dtype=float) / window
    tprt_convolved = np.convolve(tprt, weighting_function, 'same')
    ict_convolved = np.convolve(ict, weighting_function, 'same')
    space_convolved = np.convolve(space, weighting_function, 'same')
    print tprt
    print tprt_convolved
    tprt_convolved[0:(window - 1) / 2] = tprt_convolved[(window - 1) / 2]
    ict_convolved[0:(window - 1) / 2] = ict_convolved[(window - 1) / 2]
    space_convolved[0:(window - 1) / 2] = space_convolved[(window - 1) / 2]
    tprt_convolved[-(window - 1) / 2:] = tprt_convolved[-((window + 1) / 2)]
    ict_convolved[-(window - 1) / 2:] = ict_convolved[-((window + 1) / 2)]
    space_convolved[-(window - 1) / 2:] = space_convolved[-((window + 1) / 2)]
    print tprt_convolved
    new_tprt = np.transpose(np.tile(tprt_convolved, (columns, 1)))
    new_ict = np.transpose(np.tile(ict_convolved, (columns, 1)))
    new_space = np.transpose(np.tile(space_convolved, (columns, 1)))

    # calibrating thermal channel

    tBB = new_tprt
    tsBB = A + B * tBB
    nBB = (1.1910427 * 0.000010) * cWavenumber * cWavenumber * cWavenumber
    nBB = nBB / (np.exp((1.4387752 * cWavenumber) / tsBB) - 1.0)

    Nlin = Ns + \
        (((nBB - Ns) * (new_space - raw_counts)) / (new_space - new_ict))
    Ncor = b0 + b1 * Nlin + b2 * Nlin * Nlin
    Ne = Ncor
    tsE = 1.4387752 * cWavenumber
    tsE = tsE / \
        np.log(
            1.0 + ((1.1910427 * 0.000010 * cWavenumber * cWavenumber * cWavenumber) / Ne))
    bt = (tsE - A) / B

    return bt



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright (c) 2014 Adam.Dybbroe

# Author(s):

#   Adam.Dybbroe <a000680@c14526.ad.smhi.se>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
"""

import logging
import os

import version

__version__ = version.__version__

LOG = logging.getLogger(__name__)
try:
    CONFIG_FILE = os.environ['PYGAC_CONFIG_FILE']
except KeyError:
    LOG.error('Environment variable PYGAC_CONFIG_FILE not set!')
    CONFIG_FILE = ''

if not os.path.exists(CONFIG_FILE) or not os.path.isfile(CONFIG_FILE):
    LOG.warning(str(CONFIG_FILE) + " pointed to by the environment " +
                "variable PYGAC_CONFIG_FILE is not a file or does not exist!")




#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright (c) 2012, 2014 Abhay Devasthale and Martin Raspaud

# Author(s):

#   Abhay Devasthale <abhay.devasthale@smhi.se>
#   Martin Raspaud <martin.raspaud@smhi.se>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Read a gac file.

"""


import logging

from datetime import datetime

logger = logging.getLogger("")
logger.setLevel(logging.DEBUG)
ch = logging.StreamHandler()
ch.setLevel(logging.DEBUG)

class MyFormatter(logging.Formatter):
    converter = datetime.fromtimestamp
    def formatTime(self, record, datefmt=None):
        ct = self.converter(record.created)
        if datefmt:
            s = ct.strftime(datefmt)
        else:
            t = ct.strftime("%Y-%m-%d %H:%M:%S")
            s = "%s.%03d" % (t, record.msecs)
        return s


formatter = MyFormatter('[ %(levelname)s %(name)s %(asctime)s] %(message)s')
ch.setFormatter(formatter)
logger.addHandler(ch)

def check_file_version(filename):
    with open(filename) as fdes:
        data = fdes.read(3)
    if data in ["CMS", "NSS", "UKM", "DSS"]:
        from pygac.gac_klm import main
        return main
    else:
        from pygac.gac_pod import main
        return main



if __name__ == "__main__":
    import sys
    try:
        filename = sys.argv[1]
	start_line = sys.argv[2]
	end_line = sys.argv[3] 
    except IndexError:
	print "Usage: gac_run <filename> <start scan line number> <end scan line number>"
	sys.exit(1)

    reader = check_file_version(filename)
    try:
        reader(filename, start_line, end_line)
    except ValueError:
        print "Value error"

